--[[TODO

]]--

local Hermes = LibStub("AceAddon-3.0"):GetAddon("Hermes")
local MODNAME = "HermesDisplay"
local HermesDisplay = Hermes:NewModule(MODNAME, "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')

local HermesSCT = Hermes:GetModule("HermesSCT")
local HermesMikSBT = Hermes:GetModule("HermesMikSBT")
local HermesParrot = Hermes:GetModule("HermesParrot")

--LibButtonFacade support
local LBF  = LibStub("LibButtonFacade",true)
local LBFHermes = {} 
LBFHermes.Name = "Hermes"

--ToolTip
local QTip	= LibStub("LibQTip-1.0")
local ToolTip = nil
local SpellIdFont = nil

--frames
HermesDisplay.Containers = nil
local RecycledButtons = nil
local CommStatusFrame = nil
local BUTTON_FRAME_COUNT = 1
local CONTAINER_FRAME_COUNT = 1
local LAYOUT_FRAME_MINWIDTH = 250
local LAYOUT_FRAME_MINHEIGHT = 150

--some UI related constant
local STATUS_PADDING = 2

--other
local SCAN_FREQUENCY = 0.1						--how often OnUpdate is called for buttons requiring animation/data changes such as tooltip
local SENDER_STATUS_UPDATE_FREQUENCY = 1		--how often sender status is checked
HermesDisplay.PlayerZone = nil
HermesDisplay.Senders = nil
local SenderStatusTimer = nil
local ICONSIZE_LARGE = 32
local RESIZER_SIZE = 20

--tictacs
HermesDisplay.MAX_TICTACS = 5					--how many tick tacks will be created for each button.
												--5 is a good number due to multiboxers
												--but this number can actually be increased manually here if you
												--need more tictacs for some reason

--textures and colors
local ICON_STATUS_NOTREADY = "Interface\\RAIDFRAME\\ReadyCheck-NotReady.blp"
local ICON_STATUS_READY = "Interface\\RAIDFRAME\\ReadyCheck-Ready.blp"
local ICON_STATUS_UNKNOWN = "Interface\\RAIDFRAME\\ReadyCheck-Waiting.blp"
local IMAGE_RESIZE = "Interface\\Addons\\Hermes\\Textures\\Resize.tga"
local VertexColorUnavailable = {0.5, 0.5, 0.5, 1.0}
local VertexColorDoesntExist = {1.0, 1.0, 1.0, 0.5}
local BGCOLOR_DRAGDROP_DROPTARGET = {0, 0.75, 0, 0.5}
local BGCOLOR_DRAGDROP_NOTDROPTARGET = {0, 0, 0, 0.3}
	
--drag and drop support
local DragHandler = nil
local DragPlaceHolder = nil						--currently just used as a table, but I may add fancier support to show where buttons land when dragging dropping to arrangelayout method

--------------------------------------------------------------------
-- Common methods used by Hermes sending/receiving
--------------------------------------------------------------------
function HermesDisplay:AddTracker(id, name, icon, containerName, class)
	local container = self:FindContainerByName(containerName)
	
	if(not container) then
		--rather than error out, just throw it into the default container. Although this shouldn't happen
		container = HermesDisplay.DefaultContainer
	end
	
	local trackerInfo = {
		id = id,
		name = name,
		icon = icon,
		class = class,
		senders = {}
	}
				
	local button = self:FetchButton(icon)
	button.trackerInfo = trackerInfo
	button.container = container
	button:ClearAllPoints()
	button:SetParent(container)
	local index = self:FindButtonIndex(RecycledButtons.Buttons, button)
	tremove(RecycledButtons.Buttons, index)
	tinsert(container.Buttons, button)

	self:UpdateButtonCooldowns(container, button)
	self:UpdateButtonScripts(container, button)
	self:UpdateButtonVisuals(container, button)

	self:ArrangeLayout(container)
	self:SetContainerToVirtualSize(container)
	self:UpdateButtonBorders(container, button)
	
	return container.options.name
end

function HermesDisplay:GetContainerAndButtonForTrackerId(id)
	for _, container in ipairs(HermesDisplay.Containers) do
		for _, button in ipairs(container.Buttons) do
			if(button.trackerInfo.id == id) then
				return container, button
			end
		end
	end
end

function HermesDisplay:RemoveTracker(id)
	local container, button = self:GetContainerAndButtonForTrackerId(id)
	if(container and button) then
		self:RecycleButton(container, button)
		return
	end
end

function HermesDisplay:AddSender(senderName, classEnum)
	local sender = {
		classEnum = classEnum,
		raidId = nil,
		zone = nil,
		online = true,
		connected = true,
	}
	self.Senders[senderName] = sender
	self:UpdateSenderStatus(senderName)
	return sender
end

function HermesDisplay:RemoveSender(senderName)
	--remove all spell data for this sender
	HermesDisplay:RemoveTrackersForSender(senderName)
	
	--remove sender from senders
	HermesDisplay.Senders[senderName] = nil
end

function HermesDisplay:GetSender(senderName)
	for name, sender in pairs(HermesDisplay.Senders) do
		if(name == senderName) then
			return sender
		end
	end
	
	return nil
end

function HermesDisplay:AddSenderTrackerInfo(senderName, id, duration)
	if(not self:GetSender(senderName)) then
		error("unknown sender in AddSenderSpellInfo")
	end
	
	local container, button = self:GetContainerAndButtonForTrackerId(id)
	
	if(container and button) then
		local senderIndex = self:FindSenderIndexInButton(button, senderName)
		local sender = nil
		if(not senderIndex) then
			--no trackerinfo for this sender yet
			sender = {
				senderName = senderName
			}
			tinsert(button.trackerInfo.senders, sender)
		else
			--we have tracker info already, this sender has sent spell information before
			sender = button.trackerInfo.senders[senderIndex]
			
			--update SCT only if the duration is different than current duration and greater than zero
			if(duration and duration > 0 and sender.duration ~= duration) then
				--if(HermesSCT:IsEnabled() and container.options.SCT_Integration == true) then
				if(HermesSCT:IsEnabled() and Hermes.db.profile.SCT_Integration == true) then
					HermesSCT:SendNotReadyMessage(Hermes.PlayerName, senderName, button.trackerInfo.name, button.trackerInfo.icon, button.trackerInfo.class)
				end
				
				--if(HermesMikSBT:IsEnabled() and container.options.MikSBT_Integration == true) then
				if(HermesMikSBT:IsEnabled() and Hermes.db.profile.MikSBT_Integration == true) then
					HermesMikSBT:SendNotReadyMessage(Hermes.PlayerName, senderName, button.trackerInfo.name, button.trackerInfo.icon, button.trackerInfo.class)
				end
				
				--if(HermesParrot:IsEnabled() and container.options.Parrot_Integration == true) then
				if(HermesParrot:IsEnabled() and Hermes.db.profile.Parrot_Integration == true) then
					HermesParrot:SendNotReadyMessage(Hermes.PlayerName, senderName, button.trackerInfo.name, button.trackerInfo.icon, button.trackerInfo.class)
				end
			end
		end

		sender.duration = duration
		sender.timestamp = GetTime()
			
		self:UpdateButtonCooldowns(container, button)
		self:UpdateButtonScripts(container, button)
		self:UpdateButtonVisuals(container, button)
			
		if(sender.duration) then
			button:SetScript("OnUpdate", OnButtonUpdate) -- start the OnUpdate script if not already started
		end
	end
	
	--if we made it here, we're not tracking the spell
end

function HermesDisplay:RemoveTrackersForSender(senderName)
	for _, container in ipairs(HermesDisplay.Containers) do
		for _, button in ipairs(container.Buttons) do
			local senderIndex = self:FindSenderIndexInButton(button, senderName)
			if(senderIndex) then
				tremove(button.trackerInfo.senders, senderIndex)
				
				HermesDisplay:UpdateButtonCooldowns(container, button)
				HermesDisplay:UpdateButtonScripts(container, button)
				HermesDisplay:UpdateButtonVisuals(container, button)
			end
		end
	end
end

function HermesDisplay:RemoveTrackerForSender(id, senderName)
	local container, button = self:GetContainerAndButtonForTrackerId(id)
	if(container and button) then
		local senderIndex = self:FindSenderIndexInButton(button, senderName)
		if(senderIndex) then
			tremove(button.trackerInfo.senders, senderIndex)
			HermesDisplay:UpdateButtonCooldowns(container, button)
			HermesDisplay:UpdateButtonScripts(container, button)
			HermesDisplay:UpdateButtonVisuals(container, button)
		end
	end
end

--------------------------------------------------------------------
-- Sender Helpers
--------------------------------------------------------------------
function HermesDisplay:UpdateSenderStatus(senderName)
	local sender = self.Senders[senderName]
	if(Hermes.PlayerInRaid) then
		--if we're in a raid
		sender.raidId = UnitInRaid(senderName)
		if(sender.raidId) then
			local oldZone = sender.zone
			local oldOnline = sender.online
			local oldDead = sender.dead
			
			local name, _, _, _, _, _, zone, online, dead, _, _ = GetRaidRosterInfo(sender.raidId)
			
			if(name == senderName) then
				sender.zone = zone
				sender.online = online
				sender.dead = dead
			else
				error("Unexpected sender status during UpdateSenderStatus")
			end
			
			if(oldZone ~= sender.zone or oldOnline ~= sender.online or oldDead ~= sender.dead) then
				HermesDisplay:UpdateButtonVisuals()
			end
		else
			--sender dropped from raid
			self:RemoveSender(senderName)
		end
	elseif(Hermes.PlayerInParty) then
		--if we're in a party
		if(UnitInParty(senderName)) then
			local oldDead = sender.dead
			
			sender.zone = HermesDisplay.PlayerZone --there is no API to get the zone of a player or party member so assume zones match
			sender.online = 1 --there is no API to get the online status of a player or party member so assume online
			
			sender.dead = UnitIsDead(senderName) or UnitIsGhost(senderName)
			
			if(oldDead ~= sender.dead) then
				HermesDisplay:UpdateButtonVisuals()
			end
		else
			--sender dropped from party
			self:RemoveSender(senderName)
		end
	end
end

function HermesDisplay:OnUpdateSenderStatus()
	if(HermesDisplay.Senders) then
		for senderName, senderInfo in pairs(HermesDisplay.Senders) do
			self:UpdateSenderStatus(senderName)
		end
	end
end

--------------------------------------------------------------------
-- LibButtonFacade
--------------------------------------------------------------------
function LBFHermes.SetupGroup()
	local group = LBF:Group(LBFHermes.Name, MODNAME);
	if(not Hermes.db.profile.lbf) then
		Hermes.db.profile.lbf = { skin = "Blizzard", gloss = 0, backdrop = false, colors = {} }
	end
	
	group:Skin(
		Hermes.db.profile.lbf.skin,
		Hermes.db.profile.lbf.gloss,
		Hermes.db.profile.lbf.backdrop,
		Hermes.db.profile.lbf.colors)
end

function LBFHermes.RegisterSkin(arg, skin, gloss, backdrop, group, button, colors)
	if(group and group == MODNAME) then
		Hermes.db.profile.lbf.skin = skin
		Hermes.db.profile.lbf.gloss = gloss
		Hermes.db.profile.lbf.backdrop = backdrop
		Hermes.db.profile.lbf.colors = colors
		HermesDisplay:UpdateButtonBorders(nil, nil)
	end
end

function HermesDisplay:SetupLBFGroup()
	LBFHermes.SetupGroup()
end

function HermesDisplay:UpdateButtonBorders(container, button)
	if(container) then
		if(button) then
			if(button.trackerInfo.class and button.trackerInfo.class ~= "ANY") then
				local c = Hermes:GetClassColorRGB(button.trackerInfo.class)
				if(LBF) then
					--backup the original color if not already exists
					if(not button.lbf_nvcd) then
						button.lbf_nvcd = {}
						button.lbf_nvcd.r, button.lbf_nvcd.g, button.lbf_nvcd.b, button.lbf_nvcd.a = LBF:GetNormalVertexColor(button)
					end
					
					if(container.options.ClassColoredBorders and container.options.ClassColoredBorders == true) then
						LBF:SetNormalVertexColor(button, c.r, c.g, c.b, c.a)
					else
						LBF:SetNormalVertexColor(button,
							button.lbf_nvcd.r,
							button.lbf_nvcd.g,
							button.lbf_nvcd.b,
							button.lbf_nvcd.a)
					end
				else
					if(container.options.ClassColoredBorders and container.options.ClassColoredBorders == true) then
						button.normaltexture:SetVertexColor(c.r, c.g, c.b, c.a)
					else
						button.normaltexture:SetVertexColor(
							button.nvcd.r,
							button.nvcd.g,
							button.nvcd.b,
							button.nvcd.a)
					end
				end
			end	
		else
			for _, b in ipairs(container.Buttons) do
				self:UpdateButtonBorders(container, b)
			end
		end
	else
		for _, c in ipairs(HermesDisplay.Containers) do
			self:UpdateButtonBorders(c, nil)
		end
	end
end
--------------------------------------------------------------------
-- AceAddon, Timers, WoW events
--------------------------------------------------------------------
function HermesDisplay:OnEnable()
	
end

function HermesDisplay:OnDisable()
	
end

function HermesDisplay:OnInitialize()
	self:SetupToolTipFonts()
	self:CreateDragHandler()
	self:CreateRecycledButtonFrame()
	self:CreateRecycledContainerFrame()
	self:InitializeContainers()
	self:CreateDragPlaceHolder() --make sure to create this after creating the containers
	self:SetEnabledState(true)
	HermesDisplay.Senders = {}
	
	if(LBF) then
		self:SetupLBFGroup()
		LBF:RegisterSkinCallback(LBFHermes.Name, LBFHermes.RegisterSkin, LBFHermes)
	end
end

function HermesDisplay:EnableEvents(toggle)						--enables message and event handling
	if(toggle == true) then
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("ZONE_CHANGED")
		self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
		self:RegisterEvent("ZONE_CHANGED_INDOORS")
		self:StartSenderStatusTimer()
	else
		self:UnregisterEvent("PLAYER_ENTERING_WORLD")
		self:UnregisterEvent("ZONE_CHANGED")
		self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
		self:UnregisterEvent("ZONE_CHANGED_INDOORS")
		self:KillSenderStatusTimer()
	end
end

function HermesDisplay:ResetSenders()							--wipe out all senders and button state of senders, but keeps buttons
	if(HermesDisplay.Senders) then
		for senderName, senderInfo in pairs(HermesDisplay.Senders) do
			HermesDisplay:RemoveSender(senderName)
		end

		wipe(HermesDisplay.Senders)
		HermesDisplay.Senders = nil
	end
end

function HermesDisplay:ResetTrackerState()								--resets everything to a clean slate (for example as needed due to a profile change)
	self:ResetSenders()
	for _, container in ipairs(HermesDisplay.Containers) do
		self:RecycleButtons(container)
	end
end

function HermesDisplay:Shutdown()
	HermesDisplay:EnableEvents(false)
	HermesDisplay:ResetSenders()
end

function HermesDisplay:Startup()								--initializes data
	self:UpdateContainerVisibility(false, false) --hide by default

	if(not HermesDisplay.Senders) then
		HermesDisplay.Senders = {}
	end
	self.PlayerZone = GetRealZoneText()
	HermesDisplay:EnableEvents(true)
end

function HermesDisplay:PLAYER_ENTERING_WORLD()				--used only for one time player initialization that has to happen after player is logged in.
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:ZONE_CHANGED()
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:ZONE_CHANGED_NEW_AREA()
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:ZONE_CHANGED_INDOORS()
	self.PlayerZone = GetRealZoneText()
end

function HermesDisplay:KillSenderStatusTimer()
	if(SenderStatusTimer) then
		self:CancelTimer(SenderStatusTimer, true)
		SenderStatusTimer = self:ScheduleTimer("OnUpdateSenderStatus", 0.1, nil) --convert to one shot timer
		self:CancelTimer(SenderStatusTimer, true) --cancel again, we don't care if it ran or not
		SenderStatusTimer = nil
	end
end

function HermesDisplay:StartSenderStatusTimer()
	SenderStatusTimer = self:ScheduleRepeatingTimer("OnUpdateSenderStatus", SENDER_STATUS_UPDATE_FREQUENCY, nil)
end

--------------------------------------------------------------------
-- Recycled Button Frame
--------------------------------------------------------------------
function HermesDisplay:CreateRecycledButtonFrame()
	--RecycledButtons = CreateFrame("Frame", "HermesDisplayRecycledButtons", UIParent)
	RecycledButtons = CreateFrame("Frame", nil, UIParent)
	RecycledButtons:SetPoint("CENTER")
	RecycledButtons:SetWidth(50)
	RecycledButtons:SetHeight(50)
	RecycledButtons:Hide()

	RecycledButtons:EnableMouse(false)
	RecycledButtons:SetMovable(false)
	RecycledButtons:SetToplevel(false)
	RecycledButtons.Buttons = {}
end

--------------------------------------------------------------------
-- Recycled Container Frame
--------------------------------------------------------------------
function HermesDisplay:CreateRecycledContainerFrame()
	RecycledContainers = CreateFrame("Frame", "HermesDisplayContainerButtons", UIParent)
	RecycledContainers:SetPoint("CENTER")
	RecycledContainers:SetWidth(50)
	RecycledContainers:SetHeight(50)
	RecycledContainers:Hide()

	RecycledContainers:EnableMouse(false)
	RecycledContainers:SetMovable(false)
	RecycledContainers:SetToplevel(false)
	RecycledContainers.Containers = {}
end

--------------------------------------------------------------------
-- Virtual Grid Provider
--------------------------------------------------------------------
--return absolute width and height of a cell based on number of tic tacs and tictac direction
--should only need to be called once per grid layout
function HermesDisplay:VGP_GetCellSizes(container)
	local width, height
	local ttw = 0
	local tth = 0
	local buttonsize = ICONSIZE_LARGE + container.options.padding
	
	if(container.options.NumTicTacs > 0 and container.options.TicTacDirection == false) then
		--vertical tictacs
		ttw = container.options.TicTacWidth + container.options.padding --slight tweak here post beta5: Added padding to the width, makes it more sensible. Allows for there to always be a gap between cells if this is set.
		tth = ((container.options.TicTacHeight + container.options.TicTacGap) * container.options.NumTicTacs) + container.options.TicTacButtonOffset
		height = buttonsize + tth
		--return the greater of the two, buttonsize or tictac width. This prevents button overlap
		if(ttw < buttonsize) then
			width = buttonsize
		else
			width = ttw
		end
	elseif(container.options.NumTicTacs > 0 and container.options.TicTacDirection == true) then
		--horizontal tictacs
		ttw = ((container.options.TicTacWidth + container.options.TicTacGap) * container.options.NumTicTacs) + container.options.TicTacButtonOffset
		tth = container.options.TicTacHeight + container.options.padding  --slight tweak here post beta5: Added padding to the width, makes it more sensible. Allows for there to always be a gap between cells if this is set.
		width = buttonsize + ttw
		--return the greater of the two, buttonsize or tictac height. This prevents button overlap
		if(tth < buttonsize) then
			height = buttonsize
		else
			height = tth
		end
	else
		--no tictacs
		width = buttonsize
		height = buttonsize
	end

	return width, height, ttw, tth
end

--Returns the coordines of a virtual cell at the specified column and row relative to the anchor point of the buttons in the container
--The coordinates returns should be used when determining button coordinates
-- multX and multY are based off the button anchor and make the number negative or positive depending on the anchor
function HermesDisplay:VGP_GetCellCoordinates(multY, multX, cellw, cellh, column, row)
	local left = abs((column - 1) * (cellw)) * multX
	local bottom = abs((row - 1) * (cellh)) * multY
	local right = (cellw + abs(left)) * multX
	local top = (cellh + abs(bottom)) * multY
	return left, top, right, bottom
end

function HermesDisplay:VGP_GetButtonCoordinates(container, cellw, cellh, ttw, tth, column, row)
	local anchor, multY, multX = self:GetButtonAnchor(container)
	local x, y
	--now we can the button coordinate, their exact position depends on whether tictacs are shown and their direction
	if(container.options.NumTicTacs > 0) then
		--first lets get the coordinates of the virtual cell that the button and it's tictacs will be placed  within
		local cellL, cellT, cellR, cellB = self:VGP_GetCellCoordinates(multY, multX, cellw, cellh, column, row)
		--we're showing tictacs
		if(container.options.TicTacDirection == false and container.options.TicTacSide == false) then
			--vertical, bottom
			local cellcenterX = ((column - 1) * cellw) + (cellw / 2)
			--now that we know where the center of the cell is, we can adjust the button based on it
			if(container.options.ButtonAnchor == "TOPLEFT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB - container.options.padding
			elseif(container.options.ButtonAnchor == "TOPRIGHT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB - container.options.padding
			elseif(container.options.ButtonAnchor == "BOTTOMLEFT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB + tth + container.options.padding
			elseif(container.options.ButtonAnchor == "BOTTOMRIGHT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB + tth + container.options.padding
			end
		elseif(container.options.TicTacDirection == false and container.options.TicTacSide == true) then
			--vertical, top
			local cellcenterX = ((column - 1) * cellw) + (cellw / 2)
			if(container.options.ButtonAnchor == "TOPLEFT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB - tth - container.options.padding
			elseif(container.options.ButtonAnchor == "TOPRIGHT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB - tth - container.options.padding
			elseif(container.options.ButtonAnchor == "BOTTOMLEFT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB + container.options.padding
			elseif(container.options.ButtonAnchor == "BOTTOMRIGHT") then
				x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
				y = cellB + container.options.padding
			end
		elseif(container.options.TicTacDirection == true and container.options.TicTacSide == false) then
			--horizontal, right
			local cellcenterY = ((row - 1) * cellh) + (cellh / 2)
			--now that we know where the center of the cell is, we can adjust the button based on it
			if(container.options.ButtonAnchor == "TOPLEFT") then
				x = ((cellL + container.options.padding) * multX)
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			elseif(container.options.ButtonAnchor == "TOPRIGHT") then
				x = ((cellR + ICONSIZE_LARGE + container.options.padding))
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			elseif(container.options.ButtonAnchor == "BOTTOMLEFT") then
				x = ((cellL + container.options.padding) * multX)
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			elseif(container.options.ButtonAnchor == "BOTTOMRIGHT") then
				x = ((cellR + ICONSIZE_LARGE + container.options.padding))
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			end
		elseif(container.options.TicTacDirection == true and container.options.TicTacSide == true) then
			--horizontal, left
			local cellcenterY = ((row - 1) * cellh) + (cellh / 2)
			--now that we know where the center of the cell is, we can adjust the button based on it
			if(container.options.ButtonAnchor == "TOPLEFT") then
				x = ((cellL + ttw + container.options.padding) * multX)
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			elseif(container.options.ButtonAnchor == "TOPRIGHT") then
				x = ((cellR + ttw + ICONSIZE_LARGE + container.options.padding))
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			elseif(container.options.ButtonAnchor == "BOTTOMLEFT") then
				x = ((cellL + ttw + container.options.padding) * multX)
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			elseif(container.options.ButtonAnchor == "BOTTOMRIGHT") then
				x = ((cellR + ttw + ICONSIZE_LARGE + container.options.padding))
				y = ((cellcenterY - (ICONSIZE_LARGE / 2)) * multY)
			end
		end
	else
		local cellcenterX = ((column - 1) * cellw) + (cellw / 2)
		local cellcenterY = ((row - 1) * cellh) + (cellh / 2)

		x = (cellcenterX - (ICONSIZE_LARGE / 2)) * multX
		y = (cellcenterY - (ICONSIZE_LARGE / 2)) * multY
	end

	return anchor, x, y
end

--------------------------------------------------------------------
-- Container Methods
--------------------------------------------------------------------
function HermesDisplay:SetPositionRelativeToButtonAnchor(container, x, y)
	--this is a matter of setting the window topleft corner relative to the anchor
	if(container.options.ButtonAnchor == "TOPLEFT") then
		--just set x and y
		container.options.x = x
		container.options.y = y
	elseif(container.options.ButtonAnchor == "TOPRIGHT") then
		local deltaX = container:GetRight() - x
		local deltaY = container:GetTop() - y
		container.options.x = container.options.x - deltaX
		container.options.y = container.options.y - deltaY
	elseif(container.options.ButtonAnchor == "BOTTOMLEFT") then
		local deltaX = container:GetLeft() - x
		local deltaY = container:GetBottom() - y
		container.options.x = container.options.x - deltaX
		container.options.y = container.options.y - deltaY
	elseif(container.options.ButtonAnchor == "BOTTOMRIGHT") then
		local deltaX = container:GetRight() - x
		local deltaY = container:GetBottom() - y
		container.options.x = container.options.x - deltaX
		container.options.y = container.options.y - deltaY
	end
	
	self:RestoreContainerPosition(container) --nothing fancy required, it's already top left
	self:ArrangeLayout(container)
end

function HermesDisplay:GetPositionRelativeToButtonAnchor(container)
	if(container.options.ButtonAnchor == "TOPLEFT") then
		return container:GetLeft(), container:GetTop()
	elseif(container.options.ButtonAnchor == "TOPRIGHT") then
		return container:GetRight(), container:GetTop()
	elseif(container.options.ButtonAnchor == "BOTTOMLEFT") then
		return container:GetLeft(), container:GetBottom()
	elseif(container.options.ButtonAnchor == "BOTTOMRIGHT") then
		return container:GetRight(), container:GetBottom()
	end
end

function HermesDisplay:UpdateContainerVisibility(HermesEnabled, isReceiving)
	for _, container in ipairs(HermesDisplay.Containers) do
		self:ToggleContainerVisibility(container, HermesEnabled, isReceiving)
	end
end

function HermesDisplay:FindContainerByName(containerName)
	if(not containerName) then
		return HermesDisplay.DefaultContainer
	else
		for _, container in ipairs(HermesDisplay.Containers) do
			if(container.options.name and container.options.name == containerName) then
				return container
			end
		end
	end
end

function HermesDisplay:ToggleContainerVisibility(container, hermesEnabled, isReceiving)
	if(Hermes.db.profile.configMode == true and hermesEnabled == true) then --always show if in configMode and Hermes is enabled
		container:Show()
		return
	end
	
	if(hermesEnabled == false) then --always hide if not enabled, regardless of anything else
		container:Hide()
		return
	end
	
	if(isReceiving == true) then		--show if receiving
		container:Show()
	else									--hide otherwise
		container:Hide()
	end
end

function HermesDisplay:AddContainer(db)
	local container = self:FetchContainer()
	
	local index = self:FindContainerIndex(RecycledContainers.Containers, container)
	tremove(RecycledContainers.Containers, index)
	tinsert(HermesDisplay.Containers, container)
	
	self:InitializeContainer(container, db)
	self:ToggleContainerVisibility(container, Hermes.db.profile.enabled, Hermes:IsReceiving())

	return container
end

function HermesDisplay:InitializeContainers()
	if(not HermesDisplay.Containers) then
		HermesDisplay.Containers = {}
	end
	local dbcontainers = Hermes.db.profile.containers
	for i, db in ipairs(dbcontainers) do
		local container = self:AddContainer(db)
		if(not container.options.name) then
			HermesDisplay.DefaultContainer = container --store handy reference to default container
		end
	end
end

function HermesDisplay:InitializeContainer(container, db)
	container.options = db
	container:SetParent(UIParent)
	container:SetScale(db.scale)
	container:SetMinResize(ICONSIZE_LARGE, ICONSIZE_LARGE)

	self:RestoreContainerPosition(container)
	self:UpdateContainerResizerPosition(container)
	self:UpdateContainerLocking(container)
end

function HermesDisplay:FindContainerIndex(containers, container)
	for index, i in ipairs(containers) do
		if(i.id == container.id) then
			return index
		end
	end

	return nil
end

function HermesDisplay:FetchContainer()
	local container
	if(RecycledContainers.Containers and #RecycledContainers.Containers > 0) then
		container = RecycledContainers.Containers[1]
	else
		container = self:CreateContainer()
	end
	
	--wipe out the button table if exists
	if(container.Buttons) then
		wipe(container.Buttons)
		container.Buttons = {}
	end
	
	return container
end

function HermesDisplay:CreateContainer()
	--local container = CreateFrame("Frame", "HermesDisplayContainer"..tostring(CONTAINER_FRAME_COUNT), UIParent)
	local container = CreateFrame("Frame", nil, UIParent)
	container.id = CONTAINER_FRAME_COUNT
	container:Hide()
	container:EnableMouse(true)
	container:SetMovable(true)
	container:SetResizable(true)
	container:SetUserPlaced(false)
	container:SetToplevel(false)
	container:SetWidth(LAYOUT_FRAME_MINWIDTH)
	container:SetHeight(LAYOUT_FRAME_MINHEIGHT)
	container:RegisterForDrag("LeftButton");
	container:SetScript("OnDragStart", OnContainerDragStart)
	container:SetScript("OnDragStop", OnContainerDragStop)
	container:SetScript("OnSizeChanged", OnContainerSizeChanged)
	container:SetScript("OnHide", function() container:StopMovingOrSizing() end) --this is a precautionary measure to HOPEFULLY avoid any frames stuck to cursor issues while dragging if the frame gets hidden while dragging it (due to combat events)
	container:SetBackdrop({
			bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", 
			tile = false,
			tileSize = 32,
			edgeSize = 10,
			insets = {
				left = -1,
				right = -1,
				top = -1,
				bottom = -1
			}
		})
	container.Buttons = {}
	container:Hide()
	
	--container.resizer = CreateFrame("Frame", container:GetName().."Resizer", container)
	container.resizer = CreateFrame("Frame", nil, container)
	container.resizer:SetWidth(RESIZER_SIZE)
	container.resizer:SetHeight(RESIZER_SIZE)
	container.resizer:EnableMouse(true)
	container.resizer:RegisterForDrag("LeftButton")
	container.resizer:SetPoint("BOTTOMRIGHT")
	container.resizer:SetScript("OnMouseDown", function() container.resizer:StopMovingOrSizing() container:StopMovingOrSizing() container:StartSizing() end)
	container.resizer:SetScript("OnMouseUp", function() container.resizer:StopMovingOrSizing() container:StopMovingOrSizing() HermesDisplay:SetContainerToVirtualSize(container) end)
	container.resizer.texture = container.resizer:CreateTexture()
	container.resizer.texture:SetTexture(IMAGE_RESIZE)
	container.resizer.texture:SetDrawLayer("OVERLAY")
	container.resizer.texture:SetAllPoints()

	tinsert(RecycledContainers.Containers, container)
	CONTAINER_FRAME_COUNT = CONTAINER_FRAME_COUNT + 1
	
	return container
end

function HermesDisplay:RecycleContainers()
	while #HermesDisplay.Containers > 0 do
		self:RecycleContainer(HermesDisplay.Containers[1])
	end
end

function HermesDisplay:RecycleContainer(container)
	HermesDisplay:RecycleButtons(container)
	local index = self:FindContainerIndex(HermesDisplay.Containers, container)
	tremove(HermesDisplay.Containers, index)
	tinsert(RecycledContainers.Containers, container)
	container:SetParent(RecycledContainers)
	container:ClearAllPoints()
	container.options = nil
	if(container.Buttons) then
		wipe(container.Buttons)
		container.Buttons = {}
	end
	container.VirtualWidth = LAYOUT_FRAME_MINWIDTH
	container.VirtualHeight = LAYOUT_FRAME_MINHEIGHT
end

function HermesDisplay:UpdateContainerLocking(container)
	if(container) then
		if(Hermes.db.profile.locked == false or Hermes.db.profile.configMode == true) then
			container:SetBackdropColor(
				0,
				0,
				0,
				0.3)
			container:EnableMouse(true)
			container.resizer:EnableMouse(true)
			container.resizer:Show()
		else
			container:SetBackdropColor(
				0,
				0,
				0,
				0)
			container:EnableMouse(false) --allow to click through the frame with mouse
			container.resizer:EnableMouse(false)
			container.resizer:Hide()
		end
	else
		for _, c in ipairs(HermesDisplay.Containers) do
			self:UpdateContainerLocking(c)
		end
	end
end

function OnContainerDragStart(container, button)
	if(Hermes.db.profile.locked == false or Hermes.db.profile.configMode == true) then
		container:StartMoving()
	end
end

function OnContainerDragStop(container, button)
	container:StopMovingOrSizing()
	container.resizer:StopMovingOrSizing()
	HermesDisplay:SaveContainerPosition(container)
end

function OnContainerSizeChanged(container, width, height)
	HermesDisplay:ArrangeLayout(container)
	--HermesDisplay:SaveContainerPosition(container) --not sure if this is needed or not! we'll find out soon enough!
end

--This method was created for ONE reason, to change the size of the window WITHOUT MOVING IT!
--This was necessary to support efficient drag and drop. Without it, the frame could be too small and it
--also prevents the gripper from being covered up by buttons. It was essential
function HermesDisplay:SetContainerToVirtualSize(container)
	local left = container:GetLeft()
	local top = container:GetTop()
	local right = container:GetRight()
	local bottom = container:GetBottom()
	
	container:ClearAllPoints()
	if container.options.ButtonAnchor == "TOPLEFT" then
		container:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", left, top)
		container:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMLEFT", left + container.VirtualWidth, top - container.VirtualHeight)

	elseif container.options.ButtonAnchor == "TOPRIGHT" then
		container:SetPoint("TOPRIGHT", UIParent, "BOTTOMLEFT", right, top)
		container:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", right - container.VirtualWidth, top - container.VirtualHeight)
		
	elseif container.options.ButtonAnchor == "BOTTOMLEFT" then
		container:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", left, bottom)
		container:SetPoint("TOPRIGHT", UIParent, "BOTTOMLEFT", left + container.VirtualWidth, bottom + container.VirtualHeight)

	elseif container.options.ButtonAnchor == "BOTTOMRIGHT" then
		container:SetPoint("BOTTOMRIGHT", UIParent, "BOTTOMLEFT", right, bottom)
		container:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", right - container.VirtualWidth, bottom + container.VirtualHeight)
	end

	container:SetHeight(container.VirtualHeight) --this is VERY important. You'd think it would be obsolete but without it the position saving detects the wrong width and height. No doubt a consequence on me trying to keep the window from moving 
	container:SetWidth(container.VirtualWidth) --this is VERY important. You'd think it would be obsolete but without it the position saving detects the wrong width and height. No doubt a consequence on me trying to keep the window from moving 

	HermesDisplay:SaveContainerPosition(container)
end

function HermesDisplay:UpdateContainerResizerPosition(container)
	local anchor

	if container.options.ButtonAnchor == "TOPLEFT" then --good
		anchor = "BOTTOMRIGHT"
		RotateTexture(container.resizer.texture, 0)
		
	elseif container.options.ButtonAnchor == "TOPRIGHT" then --good
		anchor = "BOTTOMLEFT"
		RotateTexture(container.resizer.texture, -90)
		
	elseif container.options.ButtonAnchor == "BOTTOMLEFT" then --good
		anchor = "TOPRIGHT"
		RotateTexture(container.resizer.texture, 90)
		
	elseif container.options.ButtonAnchor == "BOTTOMRIGHT" then
		anchor = "TOPLEFT"
		RotateTexture(container.resizer.texture, 180)
		
	else
		anchor = "BOTTOMRIGHT"
		RotateTexture(container.resizer.texture, 0)
		
	end

	
	container.resizer:ClearAllPoints()
	container.resizer:SetPoint(anchor)
	container.resizer:SetScript("OnMouseDown", function() container.resizer:StopMovingOrSizing() container:StopMovingOrSizing() container:StartSizing(anchor) end)
end

function HermesDisplay:ArrangeLayout(container, dragging)
	local cellw, cellh, ttw, tth = self:VGP_GetCellSizes(container)

	table.sort(container.Buttons,
		function(a, b)
			return a.trackerInfo.name < b.trackerInfo.name
		end
	)

	local columnsPerRow = floor(container:GetWidth() / cellw)

	if(columnsPerRow <= 0) then
		columnsPerRow = 1
	end
	
	--these values are used for determining virtual layout size
	local maxRows = 1 --always make sure there is at least 1, even if there isn't
	local maxColumns = 1 --always make sure there is at least 1, even if there isn't
	
	local row = 1
	local column = 1
	
	for i, button in ipairs(container.Buttons) do
		if(button.hide == true) then
			if(button:IsVisible()) then
				button:Hide()
				button:ClearAllPoints()
			end
		else
			local anchor, x, y = self:VGP_GetButtonCoordinates(container, cellw, cellh, ttw, tth, column, row)
			self:PositionButton(container, button, anchor, x, y)
			if(container.options.NumTicTacs > 0) then
				if(container.options.TicTacDirection == false) then
					self:LayoutTicTacsVertical(button, dragging)
				else
					self:LayoutTicTacsHorizontal(button, dragging)
				end
				--show tictacs
				self:ShowTicTacs(button, true)
			else
				--hide tictacs
				self:ShowTicTacs(button, false)
			end
			
			if(i < #container.Buttons) then
				column = column + 1
				if(column == columnsPerRow + 1) then
					row = row + 1
					column = 1
				end
			end
		end
		
		if row > maxRows then maxRows = row end
		if column > maxColumns then maxColumns = column end
	end

	self:UpdateContainerVirtualSizes(container, maxRows, maxColumns, ttw, tth)
	
	-- LBF ANONYMOUS SKIN SUPPORT
	if(LBF) then
		--LBF:Group(LBFHermes.Name, MODNAME):ReSkin() --resizing buttons requires reskinning...
	end
end

function HermesDisplay:UpdateContainerVirtualSizes(container, numRows, numColumns, ttw, tth)
	local width, height
	
	if(container.options.TicTacDirection == false) then
		 --vertical
		width = (numColumns * (ICONSIZE_LARGE + container.options.padding))
		height = numRows * (ICONSIZE_LARGE + container.options.padding + tth)
		if(width < ttw) then width = ttw end
	else
		--horizontal
		width = numColumns * (ICONSIZE_LARGE + container.options.padding + ttw)
		height = (numRows * (ICONSIZE_LARGE + container.options.padding))
		if(height < tth) then height = tth end
	end

	width = width + (RESIZER_SIZE / 3)
	height = height + (RESIZER_SIZE / 3)
	
	if(container:GetWidth() > width) then
		width = container:GetWidth()
	end
	if(container:GetHeight() > height) then
		height = container:GetHeight()
	end
	
	if(width <= (ICONSIZE_LARGE + RESIZER_SIZE / 3)) then
		width = ICONSIZE_LARGE + (RESIZER_SIZE / 3) --always make sure the resizer is clickable
	end
	
	if(height <= (ICONSIZE_LARGE + RESIZER_SIZE / 3)) then
		height = ICONSIZE_LARGE + (RESIZER_SIZE / 3) --always make sure the resizer is clickable
	end

	container.VirtualWidth = width
	container.VirtualHeight = height
end

function HermesDisplay:SaveContainerPosition(container)
	container.options.x = container:GetLeft()
	container.options.y = container:GetTop()
	container.options.w = container:GetWidth()
	container.options.h = container:GetHeight()
	container:SetUserPlaced(false)

	Hermes:UpdateContainerProfilePosition(container,
		container.options.x,
		container.options.y,
		container.options.w,
		container.options.h)
end

function HermesDisplay:RestoreContainerPosition(container)				--sets the starting point for all window positions based on profile settings (if it exists)
	if(not container.options.x or not container.options.y) then
		container:ClearAllPoints()
		container:SetPoint("CENTER", UIParent, "CENTER")
		container:SetWidth(LAYOUT_FRAME_MINWIDTH)
		container:SetHeight(LAYOUT_FRAME_MINHEIGHT)
		container.options.x = container:GetLeft()
		container.options.y = container:GetTop()
		container.options.w = LAYOUT_FRAME_MINWIDTH
		container.options.h = LAYOUT_FRAME_MINHEIGHT
	else
		container:ClearAllPoints()
		container:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT",
			container.options.x,
			container.options.y
		)
		container:SetWidth(container.options.w)
		container:SetHeight(container.options.h)
		self:ArrangeLayout(container)
	end

	container:SetUserPlaced(nil)
end

--------------------------------------------------------------------
-- Buttons
--------------------------------------------------------------------
function HermesDisplay:PositionButton(container, button, anchor, x, y)
	button:ClearAllPoints()
	button:SetPoint(anchor, container, anchor, x, y)
	button:SetHeight(ICONSIZE_LARGE)
	button:SetWidth(ICONSIZE_LARGE)
	button:Show()
end

function HermesDisplay:GetButtonAnchor(container)
	local buttonanchor = container.options.ButtonAnchor
	local anchor, multY, multX
	if buttonanchor == "TOPLEFT" then
		anchor = "TOPLEFT"
		multY = -1
		multX = 1
	elseif buttonanchor == "TOPRIGHT" then
		anchor = "TOPRIGHT"
		multY = -1
		multX = -1
	elseif buttonanchor == "BOTTOMLEFT" then
		anchor = "BOTTOMLEFT"
		multY = 1
		multX = 1
	elseif buttonanchor == "BOTTOMRIGHT" then
		anchor = "BOTTOMRIGHT"
		multY = 1
		multX = -1
	else
		anchor = "TOPLEFT"
		multY = -1
		multX = 1
	end

	return anchor, multY, multX
end

function HermesDisplay:CreateButton()
	--button
	local button = CreateFrame("Button", "HermesDisplayButton"..tostring(BUTTON_FRAME_COUNT), RecycledButtons, "ActionButtonTemplate")
	button.id = BUTTON_FRAME_COUNT
	button.hide = false
	button.icon = _G[button:GetName().."Icon"]
	button.flash = _G[button:GetName().."Flash"]
	button.count = _G[button:GetName().."Count"]
	button.name = _G[button:GetName().."Name"]
	button.border = _G[button:GetName().."Border"]
	button.cooldown = _G[button:GetName().."Cooldown"]
	button.normaltexture = _G[button:GetName().."NormalTexture"]
	button.pushedtexture = _G[button:GetName().."PushedTexture"]
	button.highlighttexture = _G[button:GetName().."HighlightTexture"]
	button.checkedtexture = _G[button:GetName().."CheckedTexture"]
	
	button.nvcd = {}
	button.nvcd.r, button.nvcd.g, button.nvcd.b, button.nvcd.a = button.normaltexture:GetVertexColor()
						
	button:EnableMouse(true)
	button:SetMovable(true)
	button:SetResizable(false)
	button:RegisterForClicks("AnyUp")
	button:SetScript("OnHide", function() button:StopMovingOrSizing() end) --this is a precautionary measure to HOPEFULLY avoid any frames stuck to cursor issues while dragging if the frame gets hidden while dragging it (due to combat events)
	button:SetMotionScriptsWhileDisabled(true)
	button:SetToplevel(true)
	button:SetUserPlaced(nil)
	
	--create tictacs
	for i = 1, HermesDisplay.MAX_TICTACS do
		self:CreateTicTac(button)
	end
	--Temporarily disabled to push out new build
	button:RegisterForDrag("LeftButton");
	button:SetScript("OnDragStart", OnButtonDragStart)
	button:SetScript("OnDragStop", OnButtonDragStop)
	
	tinsert(RecycledButtons.Buttons, button)
	BUTTON_FRAME_COUNT = BUTTON_FRAME_COUNT + 1

	--LBF SUPPORT
	if(LBF) then
		LBF:Group(LBFHermes.Name, MODNAME):AddButton(button)
	else
		--we don't want to see the normal texture (which acts like a border for a checkbox) if we don't have button facade running)
		--hide doesn' work because once we click the button it'll show again, but we can clear all the points so it won't show
		button.normaltexture:ClearAllPoints()
		button.normaltexture:Hide()
	end
	
	return button
end

function HermesDisplay:ResetButton(button)
	button.trackerInfo = nil
	button.LastScan = 0

	button:SetScript("OnMouseUp", nil)
	button:SetScript("OnEnter", nil)
	button:SetScript("OnLeave", nil)
	button:SetScript("OnUpdate", nil)
	
	button.icon:SetTexture("")
	button.icon:SetDesaturated(true)
		
	button.hide = false
end

function HermesDisplay:FetchButton(texture)
	local button
	if(RecycledButtons.Buttons and #RecycledButtons.Buttons > 0) then
		button = RecycledButtons.Buttons[1]
	else
		button = self:CreateButton()
	end
	
	HermesDisplay:ResetButton(button)
	
	button:SetScript("OnMouseUp", OnButtonMouseUp)
	button:SetScript("OnEnter", OnButtonEnter)
	button:SetScript("OnLeave", OnButtonLeave)
	button:SetScript("OnUpdate", nil)

	button.icon:SetTexture(texture)
	button.icon:SetDesaturated(true)
	
	return button
end

function HermesDisplay:FindButtonIndex(buttons, button)
	for index, i in ipairs(buttons) do
		if(i.id == button.id) then
			return index
		end
	end

	return nil
end

function HermesDisplay:CalculateButtonState(button)
	local totalsenders = 0
	local availableSenders = 0
	local mindurationAvailableSenders = nil
	local unavailableSenders = 0
	local mindurationUnavailableSenders = nil
	
	for i, senderInfo in ipairs(button.trackerInfo.senders) do
		totalsenders = totalsenders + 1
		local sender = self:GetSender(senderInfo.senderName);
		if(sender.online == 1 and not sender.dead and sender.zone == self.PlayerZone) then --sender is available
			if(senderInfo.duration == nil) then
				--increment available count
				availableSenders = availableSenders + 1
			else
				--if((not mindurationAvailableSenders and senderInfo.duration ~= Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED) or (senderInfo.duration ~= Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED and senderInfo.duration < mindurationAvailableSenders)) then
				if(not mindurationAvailableSenders or senderInfo.duration < mindurationAvailableSenders) then
					--set minimum duration
					mindurationAvailableSenders = senderInfo.duration
				end
			end
		elseif(sender.online == 0 or sender.dead or sender.zone ~= self.PlayerZone) then --sender not available
			if(senderInfo.duration == nil) then
				--increment unavailable count
				unavailableSenders = unavailableSenders + 1
			else
				--if((not mindurationUnavailableSenders and senderInfo.duration ~= Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED) or (senderInfo.duration ~= Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED and senderInfo.duration < mindurationUnavailableSenders)) then
				if(not mindurationUnavailableSenders or senderInfo.duration < mindurationUnavailableSenders) then
					--set minimum duration
					mindurationUnavailableSenders = senderInfo.duration
				end
			end
		end
	end
	
	return totalsenders, availableSenders, mindurationAvailableSenders, unavailableSenders, mindurationUnavailableSenders
end

function HermesDisplay:SetButtonStyleAvailable(button, availableCount, minduration, showOverlay, killtimer)
	button:SetButtonState("NORMAL", false)
	button.icon:SetDesaturated(false)
	
	if(showOverlay) then
		button.icon:SetVertexColor(
			VertexColorUnavailable[1],
			VertexColorUnavailable[2],
			VertexColorUnavailable[3],
			VertexColorUnavailable[4])
			
		button.hide = button.container.options.HideButtonNoAvailableSenders and button.container.options.HideButtonNoSenders --make sure both options are checked
	else
		button.icon:SetVertexColor(1, 1, 1, 1)
		button.hide = false
	end
	
	--this is a straight up hack. I want to make sure that any final "sparks" are shown when a cooldown finishes
	--but I also want to make sure that we stop timers appropriately too. I'm going on the assumption that sparks will
	--never last more than 1 second in duration, and when you factor in lag, update frequency, etc, I'm doubling it.
	--Therefore, if the minDuration is > 2 seconds, don't kill it, otherwise do...
	if(minduration and minduration > 2 or killtimer and killtimer == true) then
		CooldownFrame_SetTimer(button.cooldown, 0, 0, 0) --stop cooldown if it's running
	end
	
	--now update the text
	button.count:SetText(tostring(availableCount))
end

function HermesDisplay:SetButtonStyleUnavailable(button)
	button.icon:SetDesaturated(true)
	button.icon:SetVertexColor(
		VertexColorDoesntExist[1],
		VertexColorDoesntExist[2],
		VertexColorDoesntExist[3],
		VertexColorDoesntExist[4])
	button.count:SetText(nil)
	CooldownFrame_SetTimer(button.cooldown, 0, 0, 0) --stop cooldown if it's running
	button.hide = button.container.options.HideButtonNoSenders
end

function HermesDisplay:SetButtonStyleOnCooldown(button, duration, showOverlay)
	button:SetButtonState("NORMAL", false)
	button.icon:SetDesaturated(false)
	
	if(showOverlay) then
		button.icon:SetVertexColor(
			VertexColorUnavailable[1],
			VertexColorUnavailable[2],
			VertexColorUnavailable[3],
			VertexColorUnavailable[4])
	else
		button.icon:SetVertexColor(1, 1, 1, 1)
	end
	
	button.count:SetText(nil)
	
	--ensure duration is actually a value
	if(duration) then
		CooldownFrame_SetTimer(button.cooldown, GetTime(), duration, 1) --start/reset cooldown
	end
	
	button.hide = false
end

function HermesDisplay:UpdateButtonVisuals(container, button)
	if(container) then
		if(button) then
			--resort the senders by availability
			table.sort(button.trackerInfo.senders, function(a, b) return self:SortSendersByAvailability(a, b) end)
			
			local wasHidden = button.hide
			local totalsenders, availableSenders, mindurationAvailableSenders, unavailableSenders, mindurationUnavailableSenders = HermesDisplay:CalculateButtonState(button)
			if(totalsenders > 0) then
				if(availableSenders == 0) then
					--we may or may not be on cooldown, it might be that we have some unavailable senders instead
					if(mindurationAvailableSenders ~= nil) then
						--there is an active sender with a cooldown, so go ahead and start a cooldown
						HermesDisplay:SetButtonStyleOnCooldown(button, mindurationAvailableSenders, false)
					else
						--There aren't any active senders with a cooldown running, yet totalsenders is > 0.
						--This can only mean that there are inactive senders. We don't care about their status though.
						--This will reset button count to 0, create a "dark" overlay, and will reset the timer
						HermesDisplay:SetButtonStyleAvailable(button, "", nil, true, true)
					end
				else
					HermesDisplay:SetButtonStyleAvailable(button, availableSenders, mindurationAvailableSenders, false)
				end
			else
				HermesDisplay:SetButtonStyleUnavailable(button)
			end
			
			--if the button's visibility state changed, then redo layout
			if((wasHidden == true and button.hide == false) or (wasHidden == false and button.hide == true)) then
				self:ArrangeLayout(button.container)
			end
			
			--now update the tictacs
			self:UpdateTicTacVisuals(button)
		else
			for _, i in ipairs(container.Buttons) do
				self:UpdateButtonVisuals(container, i)
			end
		end
	else
		for _, container in ipairs(HermesDisplay.Containers) do
			self:UpdateButtonVisuals(container, nil)
		end
	end
end

function HermesDisplay:UpdateButtonScripts(container, button)		--resets the OnUpdate script is necessary when no cooldowns are running
	if(container) then
		if(button) then
			for i, senderInfo in ipairs(button.trackerInfo.senders) do
				--TODO: Here would be a good place to update a tictac for each sender if time is added.
				if(senderInfo.duration) then
					return
				end
			end
			button:SetScript("OnUpdate", nil)
		else
			for _, b in ipairs(container.Buttons) do
				self:UpdateButtonScripts(container, b)
			end
		end
	else
		for _, c in ipairs(HermesDisplay.Containers) do
			self:UpdateButtonScripts(c, nil)
		end
	end
end

function HermesDisplay:GetIconState(duration, dead, connected, zone)
	if(dead or not connected or zone ~= self.PlayerZone) then
		return ICON_STATUS_NOTREADY, true --show icon
	else
		if(duration) then
			return ICON_STATUS_NOTREADY, false --don't show icon
		else
			return ICON_STATUS_READY, true --show icon
		end
	end
end

function HermesDisplay:UpdateButtonCooldowns(container, button)	--this method is called to decrement the duration of each sender since it was last called
	if(container) then
		if(button) then
			local updateButtonVisuals = false
			for i, senderInfo in ipairs(button.trackerInfo.senders) do
				--if the sender has a cooldown value (counting down)
				if(senderInfo.duration) then
					--update the cooldown
					local elapsed = GetTime() - senderInfo.timestamp
					senderInfo.duration = senderInfo.duration - elapsed
					senderInfo.timestamp = GetTime() --set the timestamp too
					--if cooldown expired, then update the button
					if(senderInfo.duration <= 0) then
						senderInfo.duration = nil --reset sender cooldown
						updateButtonVisuals = true
						
						--if(HermesSCT:IsEnabled() and container.options.SCT_Integration == true) then
						if(HermesSCT:IsEnabled() and Hermes.db.profile.SCT_Integration == true) then
							HermesSCT:SendReadyMessage(Hermes.PlayerName, senderInfo.senderName, button.trackerInfo.name, button.trackerInfo.icon, button.trackerInfo.class)
						end
						
						--if(HermesMikSBT:IsEnabled() and container.options.MikSBT_Integration == true) then
						if(HermesMikSBT:IsEnabled() and Hermes.db.profile.MikSBT_Integration == true) then
							HermesMikSBT:SendReadyMessage(Hermes.PlayerName, senderInfo.senderName, button.trackerInfo.name, button.trackerInfo.icon, button.trackerInfo.class)
						end
						
						--if(HermesParrot:IsEnabled() and container.options.Parrot_Integration == true) then
						if(HermesParrot:IsEnabled() and Hermes.db.profile.Parrot_Integration == true) then
							HermesParrot:SendReadyMessage(Hermes.PlayerName, senderInfo.senderName, button.trackerInfo.name, button.trackerInfo.icon, button.trackerInfo.class)
						end
					end
				end
				
				--now let's update the tictac timers as applicable, but only for the max amount of tictacs. There might be more senders than tictacs
				if(button.tictacs and i <= #button.tictacs) then
					local sender = self:GetSender(senderInfo.senderName)
					local icon, showIcon = self:GetIconState(senderInfo.duration, sender.dead, sender.connected, sender.zone)
					HermesDisplay:UpdateTicTacDurationState(button.tictacs[i],
						showIcon,
						icon,
						senderInfo.duration)
				end
			end
			
			if(updateButtonVisuals == true) then
				self:UpdateButtonVisuals(container, button) --update button to reflect that cooldown is finished for one or more senders
			end
		else
			for _, b in ipairs(container.Buttons) do
				self:UpdateButtonCooldowns(container, b)
			end
		end
	else
		for _, c in ipairs(HermesDisplay.Containers) do
			self:UpdateButtonCooldowns(c, nil)
		end
	end
end

function OnButtonUpdate(button, elapsed)
	button.LastScan = button.LastScan + elapsed;
	while (button.LastScan > SCAN_FREQUENCY) do
		button.LastScan = button.LastScan - SCAN_FREQUENCY
		HermesDisplay:UpdateButtonCooldowns(button.container, button)
		HermesDisplay:UpdateButtonScripts(button.container, button)
		if(ToolTip and ToolTip.button and ToolTip.button == button) then
			--refresh the tooltip since it's being displayed for this button
			HermesDisplay:RefreshTooltip()
		end
	end
end

function OnButtonEnter(button, mousebutton)
	if(not button:IsDragging()) then --don't show the tooltip if we're dragging
		HermesDisplay:ShowTooltip(button, button)
	end
end

function OnButtonLeave(button, mousebutton)
	QTip:Release(ToolTip)
	ToolTip = nil
end

function OnButtonMouseUp(button, mousebutton)
	--resync
	if(mousebutton == "LeftButton" and IsShiftKeyDown()) then
		wipe(button.trackerInfo.senders)
		HermesDisplay:UpdateButtonCooldowns(button.container, button)
		HermesDisplay:UpdateButtonScripts(button.container, button)
		HermesDisplay:UpdateButtonVisuals(button.container, button)
		if(Hermes:IsReceiving()) then --prevent trying to send messages when not in raid or party. Otherwise Lua error
			Hermes:RequestTrackerUpdate(button.trackerInfo.id)
		end
	end
	
	--toggle between containers
	if(mousebutton == "MiddleButton") then
		local currentName, priorName = Hermes:GetTrackerContainerNames(button.trackerInfo.id)
		if(priorName ~= currentName) then
			local priorContainer = HermesDisplay:FindContainerByName(priorName)
			HermesDisplay:MoveButtonBetweenContainers(button, priorContainer)
		end
	end
end

function HermesDisplay:FindSenderIndexInButton(button, senderName)
	for i, sender in ipairs(button.trackerInfo.senders) do
		if(sender.senderName == senderName) then
			return i
		end
	end
	
	return nil
end

function HermesDisplay:RecycleButton(container, button)
	local index = self:FindButtonIndex(container.Buttons, button)
	tremove(container.Buttons, index)
	tinsert(RecycledButtons.Buttons, button)
	self:ResetButton(button)
	button:SetParent(RecycledButtons)
	button:SetAllPoints()
	button.container = nil
	self:ArrangeLayout(container)
end

function HermesDisplay:RecycleButtons(container)
	if container.Buttons then
		while #container.Buttons > 0 do
			self:RecycleButton(container, container.Buttons[1])
		end
	end
end

--------------------------------------------------------------------
-- Tic-Tacs
--------------------------------------------------------------------
function HermesDisplay:CreateTicTac(button)
	if(not button.tictacs) then
		button.tictacs = {}
	end
	--tictac frame
	--local tictac = CreateFrame("Frame", button:GetName()..tostring(#button.tictacs + 1), button)
	local tictac = CreateFrame("Frame", nil, button)
	tictac.id = #button.tictacs + 1
	tictac:EnableMouse(false) --allow clickthrough
	tictac:SetMovable(false)
	tictac:SetResizable(false)
	tictac:SetToplevel(true)
	
	--tictac color
	tictac.color = tictac:CreateTexture(nil, "BORDER")
	tictac.color:SetBlendMode("DISABLE")
	tictac.color:SetTexture(1, 1, 1, 0.5)
	tictac.color:SetGradientAlpha("VERTICAL", 0, 0, 0, 0.5, 1, 1, 1, 0)
	tictac.color:SetAllPoints()
	
	--tictac status
	tictac.status = tictac:CreateTexture(nil, "OVERLAY")
	tictac.status:SetTexture("")

	--tictac player
	tictac.player = tictac:CreateFontString(nil, "ARTWORK", "GameFontWhite")
	tictac.player:SetWordWrap(false)
	tictac.player:SetJustifyV("CENTER");
	tictac.player:SetJustifyH("LEFT");
	tictac.player:SetText("")
	tictac.player:SetTextColor(0, 0, 0, 0)
	tictac.player:SetPoint("RIGHT")
	
	--tictac duration timer
	tictac.durationtimer = tictac:CreateFontString(nil, "ARTWORK", "GameFontWhite")
	tictac.durationtimer:SetWordWrap(false)
	tictac.durationtimer:SetJustifyV("CENTER");
	tictac.durationtimer:SetJustifyH("RIGHT");
	tictac.durationtimer:SetText("")
	tictac.durationtimer:SetTextColor(0, 0, 0, 0)
	tictac.durationtimer:SetPoint("RIGHT")
	
	tinsert(button.tictacs, tictac)
	return tictac
end

function HermesDisplay:ShowTicTacs(button, show)
	if(button.tictacs) then
		if(show) then
			local cnt = 0
			for _, tt in ipairs(button.tictacs) do
				--here is where we control just how many tick tacks are shown based on db value
				if(cnt < button.container.options.NumTicTacs) then
					tt:Show()
				else
					tt:Hide()
				end
				cnt =  cnt + 1
			end
		else
			for _, tt in ipairs(button.tictacs) do
				tt:Hide()
			end
		end
	end
end

function HermesDisplay:LayoutTicTacsVertical(button, dragging)
	for cnt = 1, #button.tictacs do
		local tictac = button.tictacs[cnt]
		local offsety, width, height, anchorTicTac, anchorButton
		
		if(button.container.options.TicTacSide == false) then --down
			offsety = ((( cnt - 1 ) * (button.container.options.TicTacHeight + button.container.options.TicTacGap)) * -1)
			offsety = offsety - button.container.options.TicTacButtonOffset -- new
			anchorButton = "BOTTOM"
			anchorTicTac = "TOP"
			width = button.container.options.TicTacWidth
			height = button.container.options.TicTacHeight
		else --up
			anchorButton = "TOP"
			anchorTicTac = "BOTTOM"
			offsety = ( cnt - 1 ) * (button.container.options.TicTacHeight + button.container.options.TicTacGap)
			offsety = offsety + button.container.options.TicTacButtonOffset -- new
			width = button.container.options.TicTacWidth
			height = button.container.options.TicTacHeight
		end

		tictac:ClearAllPoints()
		tictac:SetPoint(anchorTicTac, button, anchorButton, 0, offsety)
		tictac:SetHeight(height)
		tictac:SetWidth(width)
		
		tictac.status:ClearAllPoints()
		tictac.status:SetPoint("RIGHT", -STATUS_PADDING, 0)
		tictac.status:SetHeight(tictac:GetHeight() * button.container.options.StatusIconSizeMultiplier)
		tictac.status:SetWidth(tictac:GetHeight() * button.container.options.StatusIconSizeMultiplier)
		
		tictac.player:ClearAllPoints()
		tictac.player:SetFont(select(1, tictac.player:GetFont()), tictac:GetHeight() * button.container.options.PlayerNameSizeMultiplier)
		tictac.player:SetPoint("LEFT", tictac, "LEFT", STATUS_PADDING, 0)
		tictac.player:SetWidth(button.container.options.PlayerNameWidth)
		
		tictac.durationtimer:ClearAllPoints()
		tictac.durationtimer:SetFont(select(1, tictac.player:GetFont()), tictac:GetHeight() * button.container.options.PlayerNameSizeMultiplier)
		tictac.durationtimer:SetPoint("RIGHT", -STATUS_PADDING, 0)
		tictac.durationtimer:SetPoint("LEFT", tictac.player, "RIGHT", 2)
	end
end

function HermesDisplay:LayoutTicTacsHorizontal(button, dragging)
	for cnt = 1, #button.tictacs do
		local tictac = button.tictacs[cnt]
		local offsetx, width, height, anchorTicTac, anchorButton
		
		if(button.container.options.TicTacSide == false) then --right
			anchorButton = "RIGHT"
			anchorTicTac = "LEFT"
			offsetx = (( cnt - 1 ) * (button.container.options.TicTacWidth + button.container.options.TicTacGap))
			offsetx = offsetx + button.container.options.TicTacButtonOffset -- new
			width = button.container.options.TicTacWidth
			height = button.container.options.TicTacHeight
		else --left
			anchorButton = "LEFT"
			anchorTicTac = "RIGHT"
			offsetx = (( cnt - 1) * (button.container.options.TicTacWidth + button.container.options.TicTacGap)) * -1
			offsetx = offsetx - button.container.options.TicTacButtonOffset -- new
			width = button.container.options.TicTacWidth
			height = button.container.options.TicTacHeight
		end

		tictac:ClearAllPoints()
		tictac:SetPoint(anchorTicTac, button, anchorButton, offsetx, 0)
		tictac:SetHeight(height)
		tictac:SetWidth(width)
		
		tictac.status:ClearAllPoints()
		tictac.status:SetPoint("RIGHT", -STATUS_PADDING, 0)
		tictac.status:SetHeight(tictac:GetHeight() * button.container.options.StatusIconSizeMultiplier)
		tictac.status:SetWidth(tictac:GetHeight() * button.container.options.StatusIconSizeMultiplier)
		
		tictac.player:ClearAllPoints()
		tictac.player:SetFont(select(1, tictac.player:GetFont()), tictac:GetHeight() * button.container.options.PlayerNameSizeMultiplier)
		tictac.player:SetPoint("LEFT", tictac, "LEFT", STATUS_PADDING, 0)
		tictac.player:SetWidth(button.container.options.PlayerNameWidth)
		
		tictac.durationtimer:ClearAllPoints()
		tictac.durationtimer:SetFont(select(1, tictac.player:GetFont()), tictac:GetHeight() * button.container.options.PlayerNameSizeMultiplier)
		tictac.durationtimer:SetPoint("RIGHT", -STATUS_PADDING, 0)
		tictac.durationtimer:SetPoint("LEFT", tictac.player, "RIGHT", 0)
	end
end

function HermesDisplay:UpdateTicTacDurationState(tictac, iconOnly, statusIcon, duration)
	if iconOnly then
		tictac.status:SetTexture(statusIcon)
		tictac.status:Show()
		tictac.durationtimer:Hide()
	else
		if duration then
			tictac.status:Hide()
			tictac.durationtimer:Show()
			tictac.durationtimer:SetText(SecondsToClock(duration))
		else
			tictac.status:Show()
			tictac.durationtimer:Hide()
			tictac.status:SetTexture(statusIcon)
			tictac.durationtimer:SetText("")
		end
	end
end

function HermesDisplay:UpdateTicTacVisuals(button)
	for i, tictac in ipairs(button.tictacs) do
		if(i <= #button.trackerInfo.senders) then
			local alpha = 1
			local senderInfo = button.trackerInfo.senders[i]
			local sender = self:GetSender(senderInfo.senderName)
			local icon, showIcon = self:GetIconState(senderInfo.duration, sender.dead, sender.connected, sender.zone)
			
			if icon == ICON_STATUS_NOTREADY then
				alpha = button.container.options.TicTacGradientAlphaNotAvailable
			end

			--tictac color
			local color = Hermes:GetClassColorRGB(sender.classEnum)
			tictac.color:SetTexture(color.r, color.g, color.b, alpha)
			tictac.color:SetGradientAlpha("VERTICAL", 0, 0, 0, button.container.options.TicTacGradientAlphaBottom, color.r, color.g, color.b, button.container.options.TicTacGradientAlphaTop)

			--tictac player
			tictac.player:SetText(senderInfo.senderName)
			tictac.player:SetTextColor(color.r, color.g, color.b, color.a)
			
			--tictac durationtimer
			HermesDisplay:UpdateTicTacDurationState(tictac,
				showIcon,
				icon,
				senderInfo.duration)
				
			tictac.durationtimer:SetTextColor(color.r, color.g, color.b, color.a)
		else
			--tictac color
			tictac.color:SetTexture(1, 1, 1, button.container.options.TicTacGradientAlphaNoPlayer)
			tictac.color:SetGradientAlpha("VERTICAL", 0, 0, 0, button.container.options.TicTacGradientAlphaBottom, 1, 1, 1, button.container.options.TicTacGradientAlphaTop)

			--tictac player
			tictac.player:SetText("")
			tictac.durationtimer:SetText("")
			tictac.durationtimer:SetTextColor(0, 0, 0, 0)
			
			--tictac status
			--tictac.status:SetTexture("")
			HermesDisplay:UpdateTicTacDurationState(tictac,
				true,
				"",
				nil)
		end
	end
end

--------------------------------------------------------------------
-- ToolTip
--------------------------------------------------------------------
local function SortByName(a, b)
	return a.senderName < b.senderName --sort by name if duration is the same
end

local function SortByDuration(a, b)
	local durationA 
	local durationB

	if(not a.duration) then
		durationA = 0
	else
		durationA = a.duration
	end
	
	if(not b.duration) then
		durationB = 0
	else
		durationB = b.duration 
	end

	if(durationA == durationB) then
		return SortByName(a, b)
	else
		return durationA < durationB --sort by duration
	end
end

local function SortByAlive(a, b)
	local senderA = HermesDisplay:GetSender(a.senderName)
	local senderB = HermesDisplay:GetSender(b.senderName)
	
	--alive
	if(not senderA.alive and senderB.alive) then
		return false --A below B
	end
	if(senderA.alive and not senderB.alive) then
		return true --A above B
	end
	if(not senderA.alive and not senderB.alive) then --they're both dead, sort by duration
		return SortByDuration(a, b)
	end
	
	--they're both alive, sort by duration
	return SortByDuration(a, b)
end

local function SortByZone(a, b)
	local senderA = HermesDisplay:GetSender(a.senderName)
	local senderB = HermesDisplay:GetSender(b.senderName)
	
	--zone
	if(senderA.zone ~= HermesDisplay.PlayerZone and senderB.zone == HermesDisplay.PlayerZone) then
		return false --A below B
	end
	if(senderA.zone == HermesDisplay.PlayerZone and senderB.zone ~= HermesDisplay.PlayerZone) then
		return true --A above B
	end
	if(senderA.zone ~= HermesDisplay.PlayerZone and senderB.zone ~= HermesDisplay.PlayerZone) then
		return SortByDuration(a, b)
	end
	
	--they're both in same zone as player, sort by duration
	return SortByAlive(a, b)
end

local function SortByConnected(a, b)
	local senderA = HermesDisplay:GetSender(a.senderName)
	local senderB = HermesDisplay:GetSender(b.senderName)
	
	--online
	if(not senderA.online and senderB.online) then
		return false --A below B
	end
	if(senderA.online and not senderB.online) then
		return true --A above B
	end
	if(not senderA.online and not senderB.online) then
		return SortByName(a, b)
	end

	--they're both connected, sort by duration
	return SortByZone(a, b)
end

function HermesDisplay:SortSendersByAvailability(a, b)
	--returning false means A below B
	--returning true means A above B
	return SortByConnected(a, b)
end

function HermesDisplay:SetupToolTipFonts()
	SpellIdFont = CreateFont("HermesSpellIdFont")
	local gfns = GameFontNormalSmall
	local filename, size, flags = gfns:GetFont()
	SpellIdFont:SetFont(filename, size, flags)
	local r, g, b, a = gfns:GetTextColor()
	local sr, sg, sb, sa = gfns:GetShadowColor()
	SpellIdFont:SetTextColor(r, g, b, a)
	SpellIdFont:SetShadowColor(sr, sg, sb, sa)
end

function HermesDisplay:RefreshTooltip()
	if(ToolTip ~= nil) then
		ToolTip:Clear()
		
		local spellLine = ToolTip:AddLine("", "", "", "", "")
		ToolTip:SetCell(spellLine, 1, ToolTip.button.trackerInfo.name, nil, "LEFT", nil, nil, 1, nil, 130, nil)
		ToolTip:SetCell(spellLine, 5, tostring(Hermes:DecodeTrackerIdentifier(ToolTip.button.trackerInfo.id)), SpellIdFont, "RIGHT", nil, nil, nil, 1, nil, nil)
		--local linecolor = Hermes:GetClassColorRGB(ToolTip.button.trackerInfo.class)
		--ToolTip:SetLineColor(spellLine,
		--	linecolor.r,
		--	linecolor.g,
		--	linecolor.b,
		--	0.5)
		ToolTip:SetLineColor(spellLine,
			1,
			1,
			1,
			0.4)
		
		if(#ToolTip.button.trackerInfo.senders == 0) then
			ToolTip:AddSeparator(1)
			local line = ToolTip:AddLine("", "", "", "", "")
			ToolTip:SetCell(line, 1, L["None found"], nil, "LEFT", 5) --set the id
		else
			ToolTip:AddLine(
				" ", --L["Player"],
				L["Time"],
				L["Alive"],
				L["Conn"],
				L["Zone"])
			ToolTip:AddSeparator(1)
		
			--table.sort(ToolTip.button.spellInfo.senders, function(a, b) return SortSendersByAvailability(a, b) end)
			
			for i, senderInfo in ipairs(ToolTip.button.trackerInfo.senders) do
				local sender = HermesDisplay:GetSender(senderInfo.senderName)
				
				--available, playername, duration, alive, zone, online
				local iconAvailable
				if(not senderInfo.duration and not sender.dead and sender.connected and sender.zone == self.PlayerZone) then
					iconAvailable = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				else
					iconAvailable = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				end
				
				local iconDead
				if(sender.dead) then
					iconDead = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				else
					iconDead = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				end
				
				local iconOnline
				if(sender.online) then
					iconOnline = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				else
					iconOnline = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				end
				
				local iconZone
				if(sender.zone == self.PlayerZone) then
					iconZone = "|T"..ICON_STATUS_READY..":0:0:0:0|t"
				else
					iconZone = "|T"..ICON_STATUS_NOTREADY..":0:0:0:0|t"
				end
				
				local durationText
				if(sender.dead or not sender.connected or sender.zone ~= self.PlayerZone) then
					--special case if the user is not connected, dead, or in wrong zone
					if(senderInfo.duration) then
						durationText = "|cFFFF0000"..SecondsToClock(senderInfo.duration).."|r" --red
					else
						durationText = "|cFFFF0000"..L["Ready"].."|r" --red
					end
				elseif(senderInfo.duration == nil) then
					durationText = "|cFF00FF00"..L["Ready"].."|r"							--green
				elseif(senderInfo.duration < 10) then
					durationText = "|cFFFFFF00"..SecondsToClock(senderInfo.duration).."|r" --yellow
				else
					durationText = "|cFFFF0000"..SecondsToClock(senderInfo.duration).."|r" --red
				end
				
				local line = ToolTip:AddLine(
					Hermes:GetColorCodedClassText(senderInfo.senderName, sender.classEnum),
					" "..durationText,
					" "..iconDead,
					" "..iconOnline,
					" "..iconZone
				)
			end
		end
	end
end

function HermesDisplay:ShowTooltip(button, anchor)
	if(ToolTip == nil) then
		ToolTip = QTip:Acquire(MODNAME.." Tooltip", 5, "LEFT", "RIGHT", "RIGHT", "RIGHT", "RIGHT", "RIGHT")
		ToolTip:ClearAllPoints()
		ToolTip:SmartAnchorTo(anchor)
		ToolTip.button = button --used by each button during on update to refresh the tip
		self:RefreshTooltip()
		ToolTip:Show()
	end
end

--------------------------------------------------------------------
-- Drag and Drop Support for buttons
--------------------------------------------------------------------
function HermesDisplay:HighlightContainers(dragging, currentTarget)
	for _, container in ipairs(HermesDisplay.Containers) do
		if(dragging) then
			if(currentTarget and container == currentTarget) then
				container:SetBackdropColor(
					BGCOLOR_DRAGDROP_DROPTARGET[1],
					BGCOLOR_DRAGDROP_DROPTARGET[2],
					BGCOLOR_DRAGDROP_DROPTARGET[3],
					BGCOLOR_DRAGDROP_DROPTARGET[4])
			else
				container:SetBackdropColor(
					BGCOLOR_DRAGDROP_NOTDROPTARGET[1],
					BGCOLOR_DRAGDROP_NOTDROPTARGET[2],
					BGCOLOR_DRAGDROP_NOTDROPTARGET[3],
					BGCOLOR_DRAGDROP_NOTDROPTARGET[4])
			end
		else
			--reset all the colors back to normal
			self:UpdateContainerLocking(container)
		end
	end
end

function HermesDisplay:CreateDragPlaceHolder()
	DragPlaceHolder = CreateFrame("Button", "HermesDisplayDragPlaceHolder", UIParent, "ActionButtonTemplate")
	DragPlaceHolder:Hide()
	DragPlaceHolder:SetAlpha(0.5)
	DragPlaceHolder.icon = _G[DragPlaceHolder:GetName().."Icon"]
	DragPlaceHolder.flash = _G[DragPlaceHolder:GetName().."Flash"]
	DragPlaceHolder.count = _G[DragPlaceHolder:GetName().."Count"]
	DragPlaceHolder.name = _G[DragPlaceHolder:GetName().."Name"]
	DragPlaceHolder.border = _G[DragPlaceHolder:GetName().."Border"]
	DragPlaceHolder.cooldown = _G[DragPlaceHolder:GetName().."Cooldown"]
	DragPlaceHolder.normaltexture = _G[DragPlaceHolder:GetName().."NormalTexture"]
	DragPlaceHolder.pushedtexture = _G[DragPlaceHolder:GetName().."PushedTexture"]
	DragPlaceHolder.highlighttexture = _G[DragPlaceHolder:GetName().."HighlightTexture"]
	DragPlaceHolder.checkedtexture = _G[DragPlaceHolder:GetName().."CheckedTexture"]
	
	DragPlaceHolder:SetPoint("CENTER")
	DragPlaceHolder:SetWidth(ICONSIZE_LARGE)
	DragPlaceHolder:SetHeight(ICONSIZE_LARGE)
	DragPlaceHolder:EnableMouse(true)
	DragPlaceHolder:SetMovable(true)
	DragPlaceHolder:SetResizable(true)
	DragPlaceHolder:SetToplevel(true)
	DragPlaceHolder:SetUserPlaced(nil)

	--LBF SUPPORT
	if(LBF) then
		LBF:Group(LBFHermes.Name, MODNAME):AddButton(DragPlaceHolder)
	end
end

function HermesDisplay:CreateDragHandler()
	DragHandler = CreateFrame("Frame")
	DragHandler.timer = 0
	DragHandler.frequency = 0.1
end

function OnButtonDragStop(button, mouseButton)
	DragHandler:SetScript("OnUpdate", nil)
	DragPlaceHolder:Hide()
	DragHandler.button = nil
	
	button.dragging = nil
	button:SetAlpha(1)
	button:StopMovingOrSizing()

	if(DragPlaceHolder.container and DragPlaceHolder.container ~= button.container) then
		HermesDisplay:MoveButtonBetweenContainers(button, DragPlaceHolder.container)
	end
	
	DragPlaceHolder.container = nil
	--update the container, it needs to refresh the opacity, etc.
	HermesDisplay:ArrangeLayout(button.container)
	
	HermesDisplay:HighlightContainers(false)
end

function OnButtonDragStart(button, mouseButton)
	if(ToolTip) then
		QTip:Release(ToolTip)
		ToolTip = nil
	end

	--highlight the containers with a greenish color so we know where to drag
	HermesDisplay:HighlightContainers(true, button.container)
	
	--upate the drag handler properties before we start dragging
	DragHandler.button = button
	DragHandler.container = button.container --always starts over the container of the button
	DragHandler:SetScript("OnUpdate", OnDragHandlerUpdate)
	button.dragging = 1 --flag the button as being detached
	button:SetAlpha(0.5)
	
	--position the drag button on top of the button being dragged
	--this funky scaling code is due to the fact that frames with different scales cause issues. This fixes it
	button:ClearAllPoints()
	local x, y = GetCursorPosition();
	local effScale = button:GetEffectiveScale()
	local offsetx = 0
	local offsety = 0
	button:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT",(x / effScale + offsetx),(y / effScale + offsety))
	button:SetWidth(ICONSIZE_LARGE)
	button:SetHeight(ICONSIZE_LARGE)

	--allow it to take mouse movement
	button:StartMoving()

	--update the container for the button being dragged so that the dragged button style paints differently
	HermesDisplay:ArrangeLayout(button.container, 1)
end

function HermesDisplay:MoveButtonBetweenContainers(button, newContainer)
	local oldContainer = button.container
	local oldContainerIndex = self:FindContainerIndex(HermesDisplay.Containers, oldContainer)
	local buttonIndex = self:FindButtonIndex(oldContainer.Buttons, button)
	
	--swap between tables
	tremove(oldContainer.Buttons, buttonIndex) --remove from old
	tinsert(newContainer.Buttons, button) --insert in new
	
	button.container = newContainer --set properties
	button:SetParent(newContainer) --set the new UI parent so that scaling and all that jazz works

	self:ArrangeLayout(oldContainer) --update the old container so that it no longer shows the button
	self:ArrangeLayout(newContainer) --update the new container so that it shows the button

	self:UpdateButtonVisuals(newContainer, button) --this call is important because properties such as button.hide are tied to the configuration properties of the container. Caling this will ensure they are updated
	
	--this will force the container frame size to match that of the values since adding the button
	self:SetContainerToVirtualSize(oldContainer)
	self:SetContainerToVirtualSize(newContainer)

	--finally, update the spell profile
	Hermes:UpdateTrackerContainer(button.trackerInfo.id, oldContainer.options.name, newContainer.options.name)
end

function OnDragHandlerUpdate(self, elapsed)
	DragHandler.timer = DragHandler.timer + elapsed
	if DragHandler.timer >= DragHandler.frequency then
		DragPlaceHolder.container = HermesDisplay:GetMouseOverContainerFrame()
		DragHandler.timer = 0
		HermesDisplay:HighlightContainers(true, DragPlaceHolder.container)
	end
end

function HermesDisplay:GetMouseOverContainerFrame()
	local mx, my = GetCursorPosition();
	for _, container in ipairs(HermesDisplay.Containers) do
		local s = container:GetEffectiveScale()
		x, y = mx/s, my/s;
		if ((x >= container:GetLeft()) and (x <= container:GetRight()) and (y >= container:GetBottom()) and (y <= container:GetTop())) then
			return container
		end
	end
	
	return nil
end


