local AddonName, Hermes = ...

local LIB_AceConfig = LibStub("AceConfig-3.0") or error("Required library AceConfig-3.0 not found")
local LIB_AceDB = LibStub("AceDB-3.0") or error("Required library AceDB-3.0 not found")
local LIB_AceConfigDialog = LibStub("AceConfigDialog-3.0")  or error("Required library AceConfigDialog-3.0 not found")
local LIB_AceConfigRegistry = LibStub("AceConfigRegistry-3.0") or error("Required library AceConfigRegistry-3.0 not found")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')

local newSpellTemplate = {
	class = L["-- Select --"],
	id = nil,
	name = nil,
}

local newItemTemplate = {
	class = L["-- Select --"],
	id = nil,
	name = nil,
}

local newContainerOptions = {
	name = nil,
	HideButtonNoSenders = false,
	HideButtonNoAvailableSenders = false,
	scale = 1,
	padding = 5,
	ButtonAnchor = "TOPLEFT",
	NumTicTacs = 0,
	TicTacDirection = true,
	TicTacSide = false,
	PlayerNameSizeMultiplier = 0.6,
	StatusIconSizeMultiplier = 0.75,
	PlayerNameWidth = 50,
	TicTacWidth = 75,
	TicTacHeight = 20,
	TicTacGap = 2,
	TicTacGradientAlphaTop = 0.75,
	TicTacGradientAlphaBottom = 0.75,
	TicTacGradientAlphaNoPlayer = 0.07,
	TicTacGradientAlphaNotAvailable = 0.6,
	TicTacButtonOffset = 0,
	ClassColoredBorders = false,
	--SCT_Integration = false,
	--Parrot_Integration = false,
	--MikSBT_Integration = false,
}

function Hermes:OpenConfigWindow()
	LIB_AceConfigDialog:Open(AddonName)
end

--sets up the default container if it or any of it's properties are missing
function Hermes:InitializeDefaultContainer()
	local default
	for _, container in ipairs(Hermes.db.profile.containers) do
		if(container.name == nil) then
			default = container
			break
		end
	end
	
	if(not default) then
		default = deepcopy(newContainerOptions)
		tinsert(Hermes.db.profile.containers, default)
	end
	
	--if default.locked == nil then default.locked = newContainerOptions.locked end
	if default.HideButtonNoSenders == nil then default.HideButtonNoSenders = newContainerOptions.HideButtonNoSenders end
	if default.HideButtonNoAvailableSenders == nil then default.HideButtonNoAvailableSenders = newContainerOptions.HideButtonNoAvailableSenders end
	if not default.scale then default.scale = newContainerOptions.scale end
	if not default.padding then default.padding = newContainerOptions.padding end
	if not default.ButtonAnchor then default.ButtonAnchor = newContainerOptions.ButtonAnchor end
	if not default.NumTicTacs then default.NumTicTacs = newContainerOptions.NumTicTacs end
	if default.TicTacDirection == nil then default.TicTacDirection = newContainerOptions.TicTacDirection end
	if default.TicTacSide == nil then default.TicTacSide = newContainerOptions.TicTacSide end
	if not default.PlayerNameSizeMultiplier then default.PlayerNameSizeMultiplier = newContainerOptions.PlayerNameSizeMultiplier end
	if not default.StatusIconSizeMultiplier then default.StatusIconSizeMultiplier = newContainerOptions.StatusIconSizeMultiplier end
	if not default.PlayerNameWidth then default.PlayerNameWidth = newContainerOptions.PlayerNameWidth end
	if not default.TicTacWidth then default.TicTacWidth = newContainerOptions.TicTacWidth end
	if not default.TicTacHeight then default.TicTacHeight = newContainerOptions.TicTacHeight end
	if not default.TicTacGap then default.TicTacGap = newContainerOptions.TicTacGap end
	if not default.TicTacGradientAlphaTop then default.TicTacGradientAlphaTop = newContainerOptions.TicTacGradientAlphaTop end
	if not default.TicTacGradientAlphaBottom then default.TicTacGradientAlphaBottom = newContainerOptions.TicTacGradientAlphaBottom end
	if not default.TicTacButtonOffset then default.TicTacButtonOffset = newContainerOptions.TicTacButtonOffset end
end

local slashCommands = {
	name = "Hermes Command Line",
	handler = Hermes,
	type = 'group',
	args = {
		config = {
			type = "execute",
			name = L["Open Configuration"],
			order = 0,
			func = function(info) Hermes:OpenConfigWindow() end,
		},
		enable = {
			type = "toggle",
			name = L["Toggle Hermes On/Off"],
			get = function() return Hermes.db.profile.enabled end,
			order = 1,
			set = "OnSetEnabled",
        },
		lock = {
			type = "toggle",
			name = L["Toggle Container Locking"],
			get = function(info) return Hermes.db.profile.locked end,
			order = 2,
			set = function(info, value)
				Hermes.db.profile.locked = value
				local d = Hermes:GetModule("HermesDisplay")
				Hermes:GetModule("HermesDisplay"):UpdateContainerLocking()
			end,
        },
    },
}

function Hermes:LoadDefaultOptions()
	local defaults = {
		profile = {
			enabled = true,
			configMode = true,
			enableparty = false,
			sender = { enabled = true },
			receiver = { enabled = true },
			spells = {},
			items = {},
			locked = false,
			lbf = {
				skin = "Blizzard",
				gloss = 0,
				backdrop = false,
				colors = {},
			},
			containers = {},
		},
	}
	self.db = LIB_AceDB:New("HermesDB", defaults)
end

--always assume that the upgrade could fail partway thru and write it in such a way
--that it could be run over and over again if needed (like if their game crashed)
function Hermes:DoDatabaseUpgrade()
	local dbp = self.db.profile
	if(dbp and dbp.display) then --this key indicates presense of beta4 or earlier
		--enabled												didn't change
		--enableparty											didn't change
		--sender.enabled										didn't change
		--receiver.enabled										didn't change

		--Handle container changes first. In beta4 it was called "display" and was a single table. In beta5 it changed to a numbered key list of tables with the default table having no name
		--display												need to migrate some values from display over to the default container
		local defaultContainer
		for _, container in ipairs(dbp.containers) do
			if(container.name == nil) then
				defaultContainer = container
				break
			end
		end
		
		if(not defaultContainer) then
			error("Database upgrade failed due to missing default container. Please report this issue")
			return
		end
		
		--note that the default container will already have matching defaults, so if not changed in the old profile we don't need to change it in the new
		if(dbp.display.locked ~= nil) then
			defaultContainer.locked = dbp.display.locked
		end
		if(dbp.display.hidespellsnosenders ~= nil) then
			defaultContainer.HideButtonNoSenders = dbp.display.hidespellsnosenders
		end
		if(dbp.display.hidespellsnoavailablesenders ~= nil) then
			defaultContainer.HideButtonNoAvailableSenders = dbp.display.hidespellsnoavailablesenders
		end
		if(dbp.display.scale ~= nil) then
			defaultContainer.scale = dbp.display.scale
		end
		if(dbp.display.padding ~= nil) then
			defaultContainer.padding = dbp.display.padding
		end
		if(dbp.display.buttonanchor ~= nil) then
			defaultContainer.ButtonAnchor = dbp.display.buttonanchor
		end
		if(dbp.display.buttonanchor ~= nil) then
			defaultContainer.ButtonAnchor = dbp.display.buttonanchor
		end
		if(dbp.display.lbf ~= nil) then
			dbp.lbf = deepcopy(dbp.display.lbf) --moved from display to root
		end
		if(dbp.display.x ~= nil) then
			defaultContainer.x = dbp.display.x
		end
		if(dbp.display.y ~= nil) then
			defaultContainer.y = dbp.display.y
		end
		if(dbp.display.w ~= nil) then
			defaultContainer.w = dbp.display.w
		end
		if(dbp.display.h ~= nil) then
			defaultContainer.h = dbp.display.h
		end
		if(dbp.display.locked ~= nil) then
			dbp.locked = dbp.display.locked --moved from display to root
		end
		
		--VertexColorUnavailable obsolete
		--VertexColorDoesntExist obsolete
		
		--receiver.spells										need to migrate values from receiver.spells to spells
		if(dbp.receiver and dbp.receiver.spells) then
			for _, trackerold in ipairs(dbp.receiver.spells) do
				if trackerold.spellid >= 0 then
					--only try to add if it doesn't already exist
					local exists = false
					for _, currenttracker in ipairs(dbp.spells) do
						if tonumber(currenttracker.id) == tonumber(trackerold.spellid) then
							exists = true
							break
						end
					end
					
					if(exists == false) then
						local tracker = {
							id = tonumber(trackerold.spellid),
							name = trackerold.spellname,
							class = trackerold.class,
							icon = trackerold.spellicon,
							container = nil,
							enabled = true, --enable by default if it existed
						}
						tinsert(dbp.spells, tracker)
					end
				else
					--it's an item
					--only try to add if it doesn't already exist
					local exists = false
					for _, currenttracker in ipairs(dbp.items) do
						if tonumber(currenttracker.id) == tonumber(trackerold.spellid) then
							exists = true
							break
						end
					end
					
					if(exists == false) then
						local tracker = {
							id = tonumber(trackerold.spellid),
							name = trackerold.spellname,
							class = trackerold.class,
							icon = trackerold.spellicon,
							container = nil,
							enabled = true, --enable by default if it existed
						}
						tinsert(dbp.items, tracker)
					end
				end
			end
			--wipe out old values from profile
			dbp.receiver.spells = nil
		end

		--sort spells
		table.sort(dbp.spells, function(a, b) return Hermes:SortProfileSpells(a, b) end)
		
		--sort items
		table.sort(dbp.items, function(a, b) return Hermes:SortProfileItems(a, b) end)

		--force new defaults to be added to hermes
		dbp.welcome = nil
		
		--wipe out old display setting so we don't do the upgrade again, this would also indicate a success
		dbp.display = nil
	end
	
	--upgrade from release 1.0b to 1.1b
	for _, container in ipairs(dbp.containers) do
	
		--add new options
		if not container.PlayerNameWidth then
			container.PlayerNameWidth = newContainerOptions.PlayerNameWidth
		end
		
		--remove old obsolete options
		if container.autohide then
			container.autohide = nil
		end
		if container.PlayerNameOffset then
			container.PlayerNameOffset = nil
		end
		if container.StatusIconOffset then
			container.StatusIconOffset = nil
		end
	end
	
	--upgrade from 1.0f to 1.0g, the main change required is handling the new config mode option
	--configMode is a new option and enabled by default, but if it's clear users already have their hermes
	--configured, then turn off configMode. We'll determine this by whether they are watching any spells or items
	--This also has the nice side effect that configMode will always be reset when they login or reloadui
	local resetConfigMode = false

	for i, spell in ipairs(dbp.spells) do
		if(spell.enabled and spell.enabled == true) then
			resetConfigMode = true
		end
	end
	
	for i, item in ipairs(dbp.items) do
		if(item.enabled and item.enabled == true) then
			resetConfigMode = true
		end
	end
	
	if resetConfigMode == true then
		dbp.configMode = false
	end
end

function Hermes:LoadBlizOptions()
	optionTables = {
		handler = Hermes,
		type="group",
		childGroups = "tab",
		name = "Hermes",
		args = {
			General = {
				name = L["General"], type="group", order = 0,
				args = {
					General = {
						type = "group",
						name = L["General"],
						inline = true,
						order = 0,
						args = {
							EnableHermes = {
								type = "toggle",
								name = L["Enable Hermes"],
								width = "normal",
								get = function(info) return self.db.profile.enabled end,
								order = 0,
								set = "OnSetEnabled",
							},
							ConfigMode = {
								type = "toggle",
								name = L["Config Mode"],
								width = "normal",
								desc = L["Allows configuration of containers and buttons while not in a group"],
								disabled = function() return self.db.profile.enabled == false end,
								get = function(info) return self.db.profile.configMode end,
								order = 1,
								set = "OnSetConfigMode",
							},
						},
					},
					Communication = {
						type = "group",
						name = L["Communication"],
						inline = true,
						order = 1,
						disabled = function() return self.db.profile.enabled == false end,
						args = {
							EnableSender = {
								type = "toggle",
								name = L["Enable Sending"],
								width = "normal",
								get = function(info) return Hermes.db.profile.sender.enabled end,
								order = 2,
								set = "OnSetEnableSender",
							},
							Spacer2A = {
								type = "description",
								name = L["Send cooldown messages to other players. No additional configuration is required for sending."],
								width = "double",
								order = 3,
							},
							Spacer2B = {
								type = "description",
								name = "",
								width = "full",
								order = 4,
							},
							EnableReceiver = {
								type = "toggle",
								name = L["Enable Receiving"],
								width = "normal",
								get = function(info) return Hermes.db.profile.receiver.enabled end,
								order = 5,
								set = "OnSetEnableReceiver",
							},
							Spacer3A = {
								type = "description",
								name = L["Receive cooldown messages from other players. The spells configured in the Spells tab will be requested from all Hermes users with Sending enabled."],
								width = "double",
								order = 6,
							},
							Spacer3B = {
								type = "description",
								name = "",
								width = "full",
								order = 7,
							},
							EnableParty = {
								type = "toggle",
								name = L["Enable Party Support"],
								get = function(info) return Hermes.db.profile.enableparty end,
								order = 8,
								set = "OnSetEnablePartySupport",
							},
							Spacer4A = {
								type = "description",
								name = L["By default, Hermes only communicates during raids. Check this box to allow party communications as well."],
								width = "double",
								order = 9,
							},
							Spacer4B = {
								type = "description",
								name = "",
								width = "full",
								order = 10,
							},
						},
					},
					CombatText = {
						disabled = function() return self.db.profile.enabled == false end,
						name = L["Combat Text Integration"],
						type = "group",
						inline = true,
						order = 3,
						args = {
							Parrot = {
								type = "toggle",
								name = "Parrot",
								width = "full",
								get = function(info) return Hermes.db.profile.Parrot_Integration end,
								order = 0,
								set =
									function(info, value)
										Hermes.db.profile.Parrot_Integration = value
									end,
							},
							MikScrollingBattleText = {
								type = "toggle",
								name = "MikScrollingBattleText",
								width = "full",
								get = function(info) return Hermes.db.profile.MikSBT_Integration end,
								order = 1,
								set =
									function(info, value)
										Hermes.db.profile.MikSBT_Integration = value
									end,
							},
							ScrollingCombatText = {
								type = "toggle",
								name = "ScrollingCombatText",
								width = "full",
								get = function(info) return Hermes.db.profile.SCT_Integration end,
								order = 2,
								set =
									function(info, value)
										Hermes.db.profile.SCT_Integration = value
									end,
							},
						},
					},
					UpgradeOptions = {
						name = L["Upgrade Options"],
						type = "group",
						inline = true,
						order = 4,
						args = {
							RefreshButton = {
								type = "execute",
								name = L["Update Spell/Item Defaults"],
								width = "normal",
								order = 1,
								desc = L["Recreates missing spell/item defaults (spells deleted or added since the last upgrade)"],
								func = function()
									Hermes:SetupNewProfileSpells()
									Hermes:SetupNewProfileItems()
								end,
							},
							Spacer1A = {
								type = "description",
								name = L["Recreates missing spell/item defaults (spells deleted or added since the last upgrade)"],
								width = "double",
								order = 2,
							},
							Spacer1B = {
								type = "description",
								name = "",
								width = "full",
								order = 3,
							},
						},
					},
				},
			},
			Spells = {
				name=L["Spells"], type="group", childGroups="tab", order = 1,
				disabled = function() return self.db.profile.enabled == false end,
				args = { }
			},
			Items = {
				name=L["Items"], type="group", childGroups="tab", order = 2,
				disabled = function() return self.db.profile.enabled == false end,
				args = { }
			},
			Containers = {
				name="Containers", type="group", childGroups="tree", order = 3,
				disabled = function() return self.db.profile.enabled == false end,
				args = {}
			},
		},
	}
	
	self:UpdateBlizOptionsTableReferences()
	LIB_AceConfig:RegisterOptionsTable(AddonName, optionTables)
	LIB_AceConfig:RegisterOptionsTable(AddonName.." ".."Command Line", slashCommands, {"hermes"})
	self.blizzOptionsFrame = LIB_AceConfigDialog:AddToBlizOptions(AddonName, AddonName)
	optionTables.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	optionTables.args.Profiles.order = 100
	LIB_AceConfigDialog:SetDefaultSize(AddonName, 700, 550)
end

function Hermes:UpdateBlizOptionsTableReferences()
	self:BlizOptionsTable_Spells()
	self:BlizOptionsTable_Items()
	self:BlizOptionsTable_Containers()
end

function Hermes:BlizOptionsTable_Spells()
	if(optionTables.args.Spells.args) then
		wipe(optionTables.args.Spells.args)
	end

	--setup basic class structure
	for classEnum, className in ipairs(Hermes.CLASS_ENUM) do
		
		local classGroup = {
			type = "group",
			inline = true,
			name = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(className), className),
			order = classEnum,
			args = { }
		}
		
		local groupHasSpells = false
		for i, spell in ipairs(self.db.profile.spells) do
			if(spell.class == className) then
				groupHasSpells = true
				classGroup.args[tostring(spell.id)] = {
					type = "group",
					inline = true,
					name = "",
					order = i,
					args = {
						name = {
							type = "toggle",
							name = "|T"..spell.icon..":0:0:0:0|t "..spell.name,
							order = 0,
							width = "double",
							desc = tostring(spell.id),
							get = function() return spell.enabled end,
							set =
								function(info, value)
									if(spell.enabled) then
										--remove it from container
										local HermesDisplay = self:GetModule("HermesDisplay")
										HermesDisplay:RemoveTracker(tonumber(spell.id))
									else
										--add it to container
										local HermesDisplay = self:GetModule("HermesDisplay")
										HermesDisplay:AddTracker(spell.id, spell.name, spell.icon, spell.container, spell.class)
										--request spell from senders
										if(Hermes:IsReceiving()) then
											Hermes:RequestTrackerUpdate(spell.id)
										end
									end
									spell.enabled = value
								end,
						},
						DeleteButton = {
							type = "execute",
							name = L["Delete"],
							width = "normal",
							order = 2,
							desc = spell.name,
							func = "OnDeleteSpell",
							confirm = function() return L["Spell will be deleted. Continue?"] end,
						}
					}
				}
			end
		end
		
		if(groupHasSpells) then
			optionTables.args.Spells.args[className] = classGroup
		end
	end
	
	optionTables.args.Spells.args.Create = {
		type="group",
		name= L["Add Spell"],
		inline = true,
		order = 0,
		args = {
			Instructions = {
				type = "description",
				name = "|cFF00FF00"..L["Drag and drop buttons between containers."].." "..L["New buttons go into the Default container."].."|r",
				width = "full",
				order = 0,
			},
			ClassSelect = {
				type = "select",
				name = L["Class"],
				order = 1,
				style = "dropdown",
				width = "normal",
				values = function()
					local values = {}
					values[L["-- Select --"]] = L["-- Select --"]
					for i, classFileName in ipairs(Hermes.CLASS_ENUM) do
						values[classFileName] = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(classFileName), classFileName)
					end
					return values
				end,
				get = function(info) return newSpellTemplate.class end,
				set = function(info, value) newSpellTemplate.class = value end,
			},
			SpellNameOrId = {
				type = "input",
				name = L["Name or ID"],
				order = 2,
				width = "normal",
				get = function(info)
					if(newSpellTemplate.name ~= nil) then
						return newSpellTemplate.name
					elseif(newSpellTemplate.id ~= nil) then
						return string.format("%.0f", newSpellTemplate.id)
					else
						return ""
					end
				end,
				set = function(info, value)
					local n = tonumber(self:Trim(value))
					local s = self:Trim(value)
					if(strlen(s) == 0) then s = nil end
					if(n and n ~= 0) then
						newSpellTemplate.id = n
						newSpellTemplate.name = nil
					else
						newSpellTemplate.id = nil
						newSpellTemplate.name = s
					end
				end,
			},
			AddSpellButton = {
				type = "execute",
				name = L["Add Spell"],
				width = "normal",
				order = 3,
				func = "OnAddSpell",
				disabled = function()
					--ensure that a class is selected and that spell evaluates to a spell id or name
					return newSpellTemplate.class == L["-- Select --"] or (not newSpellTemplate.name and not newSpellTemplate.id)
				end,
			},
		}
	}
end

function Hermes:BlizOptionsTable_Items()
	if(optionTables.args.Items.args) then
		wipe(optionTables.args.Items.args)
	end

	--setup basic class structure
	for classEnum, className in ipairs(Hermes.CLASS_ENUM) do
		
		local classGroup = {
			type = "group",
			inline = true,
			name = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(className), className),
			order = classEnum,
			args = { }
		}
		
		local groupHasItems = false
		for i, item in ipairs(self.db.profile.items) do
			if(item.class == className) then
				groupHasItems = true
				classGroup.args[tostring(item.id)] = {
					type = "group",
					inline = true,
					name = "",
					order = i,
					args = {
						name = {
							type = "toggle",
							name = "|T"..item.icon..":0:0:0:0|t "..item.name,
							order = 0,
							width = "double",
							desc = tostring(Hermes:DecodeTrackerIdentifier(item.id)),
							get = function() return item.enabled end,
							set =
								function(info, value)
									if(item.enabled) then
										--remove it from container
										local HermesDisplay = self:GetModule("HermesDisplay")
										HermesDisplay:RemoveTracker(tonumber(item.id))
									else
										--add it to container
										local HermesDisplay = self:GetModule("HermesDisplay")
										HermesDisplay:AddTracker(item.id, item.name, item.icon, item.container, item.class)
										--request spell from senders
										if(Hermes:IsReceiving()) then
											Hermes:RequestTrackerUpdate(item.id)
										end
									end
									item.enabled = value
								end,
						},
						DeleteButton = {
							type = "execute",
							name = L["Delete"],
							width = "normal",
							order = 2,
							desc = item.name,
							func = "OnDeleteItem",
							confirm = function() return L["Item will be deleted. Continue?"] end,
						}
					}
				}
			end
		end
		
		if(groupHasItems) then
			optionTables.args.Items.args[className] = classGroup
		end
	end
	
	optionTables.args.Items.args.Create = {
		type="group",
		name= " ",--L["Add Spell"],
		inline = true,
		order = 0,
		args = {
			Instructions = {
				type = "description",
				name = "|cFF00FF00"..L["Drag and drop buttons between containers."].." "..L["New buttons go into the Default container."].."|r",
				width = "full",
				order = 0,
			},
			ClassSelect = {
				type = "select",
				name = L["Class"],
				order = 1,
				style = "dropdown",
				width = "normal",
				values = function()
					local values = {}
					values[L["-- Select --"]] = L["-- Select --"]
					for i, classFileName in ipairs(Hermes.CLASS_ENUM) do
						values[classFileName] = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(classFileName), classFileName)
					end
					return values
				end,
				get = function(info) return newItemTemplate.class end,
				set = function(info, value) newItemTemplate.class = value end,
			},
			ItemNameOrId = {
				type = "input",
				name = L["Name or ID"],
				order = 2,
				width = "normal",
				get = function(info)
					if(newItemTemplate.name ~= nil) then
						return newItemTemplate.name
					elseif(newItemTemplate.id ~= nil) then
						return string.format("%.0f", newItemTemplate.id)
					else
						return ""
					end
				end,
				set = function(info, value)
					local n = tonumber(self:Trim(value))
					local s = self:Trim(value)
					if(strlen(s) == 0) then s = nil end
					if(n and n ~= 0) then
						newItemTemplate.id = n
						newItemTemplate.name = nil
					else
						newItemTemplate.id = nil
						newItemTemplate.name = s
					end
				end,
			},
			AddItemButton = {
				type = "execute",
				name = L["Add Item"],
				width = "normal",
				order = 3,
				func = "OnAddItem",
				disabled = function()
					--ensure that a class is selected and that spell evaluates to a spell id or name
					return newItemTemplate.class == L["-- Select --"] or (not newItemTemplate.name and not newItemTemplate.id)
				end,
			}
		}
	}
end

function Hermes:BlizOptionsTable_Containers()
	if(optionTables.args.Containers.args) then
		wipe(optionTables.args.Containers.args)
	end

	for k, container in ipairs(Hermes.db.profile.containers) do
		local containerName
		if(not container.name) then
			containerName = L["Default"]
		else
			containerName = container.name
		end
		optionTables.args.Containers.args[containerName] = {
			name = containerName,
			type = "group",
			childGroups = "tab",
			args = {
				Container = {
					name = L["Window"],
					type = "group",
					inline = false,
					order = 0,
					args = {
						Positioning = {
							name = L["Size and Position"],
							type = "group",
							inline = true,
							order = 1,
							args = {
								Scale = {
									type = "range",
									min = 0.1, max = 3, step = 0.01,
									name = L["Scale"],
									order = 4,
									width = "full",
									get = function(info) return container.scale end,
									set =
										function(info, value)
											container.scale = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.scale = container.scale
											c:SetScale(c.options.scale)
										end
								},
								ButtonAnchor = {
									type = "select",
									name = L["Button Anchor"],
									width = "full",
									desc = L["Changes the direction in which buttons grow."],
									get = function(info) return container.ButtonAnchor end,
									order = 6,
									style = "dropdown",
									set = 
										function(info, value)
											container.ButtonAnchor = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.ButtonAnchor = container.ButtonAnchor
											d:UpdateButtonVisuals(c, nil)
											d:UpdateContainerResizerPosition(c)
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
									values = {
										["TOPLEFT"] = L["Top Left"],
										["TOPRIGHT"] = L["Top Right"],
										["BOTTOMLEFT"] = L["Bottom Left"],
										["BOTTOMRIGHT"] = L["Bottom Right"],
									},
								},
								XPos = {
									type = "input",
									name = "X",
									width = "normal",
									desc = "X and Y indicate the coordinates of the specified anchor point.",
									get = 
										function(info)
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											local x, y = d:GetPositionRelativeToButtonAnchor(c)
											return tostring(Round(x))
										end,
									order = 8,
									validate = function(info, value)
										local text = strim(value)
										if(string.len(text) == 0) then
											return "Value required"
										end
										
										local x = tonumber(text)
										if(x == nil) then
											return "Must enter a number"
										end
										return true
									end,
									set = function(info, value)
										local d = self:GetModule("HermesDisplay")
										local c = d:FindContainerByName(container.name)
										local x, y = d:GetPositionRelativeToButtonAnchor(c)
										d:SetPositionRelativeToButtonAnchor(c, tonumber(strim(value)), y)
									end,
								},
								YPos = {
									type = "input",
									name = "Y",
									width = "normal",
									desc = "X and Y indicate the coordinates of the specified anchor point.",
									get = 
										function(info)
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											local x, y = d:GetPositionRelativeToButtonAnchor(c)
											return tostring(Round(y))
										end,
									order = 10,
									validate = function(info, value)
										local text = strim(value)
										if(string.len(text) == 0) then
											return "Value required"
										end
										
										local x = tonumber(text)
										if(x == nil) then
											return "Must enter a number"
										end
										return true
									end,
									set = function(info, value)
										local d = self:GetModule("HermesDisplay")
										local c = d:FindContainerByName(container.name)
										local x, y = d:GetPositionRelativeToButtonAnchor(c)
										d:SetPositionRelativeToButtonAnchor(c, x, tonumber(strim(value)))
									end,
								},
								PositionTip = {
									name = L["Tip: If trying to align containers with each other, make sure they use the same scale and take note of their Button Anchor setting."],
									type = "description",
									width = "full",
									order = 12,
								},
								ResetPosition = {
									type = "execute",
									name = L["Reset Position"],
									width = "full",
									desc = L["Move the container to the center of the screen."],
									order = 14,
									func =
										function(info, value)
											local HermesDisplay = self:GetModule("HermesDisplay")
											local container = HermesDisplay:FindContainerByName(container.name)
											container:ClearAllPoints()
											container:SetPoint("CENTER", UIParent, "CENTER")
											container.options.x = container:GetLeft()
											container.options.y = container:GetTop()
											HermesDisplay:SaveContainerPosition(container)
										end,
								},
							},
						},
						Other = {
							name = " ",
							type = "group",
							inline = true,
							order = 3,
							args = {
								Delete = {
									type = "execute",
									name = L["Delete"],
									width = "full",
									order = 0,
									disabled = function() return container.name == nil end,
									hidden = function() return container.name == nil end,
									func = "OnDeleteContainer",
									confirm = function() return L["Container will be deleted and its buttons placed into the Default container. Continue?"] end,
								},
							},
						},
					},
				},
				Buttons = {
					name = L["Buttons"],
					type = "group",
					inline = false,
					order = 1,
					args = {
						HideButtonNoSenders = {
							type = "toggle",
							name = L["Hide Buttons without Senders"],
							desc = L["Hides a button when no Senders have it."],
							width = "full",
							get = function(info) return container.HideButtonNoSenders end,
							order = 2,
							set =
								function(info, value)
									container.HideButtonNoSenders = value
									local d = self:GetModule("HermesDisplay")
									local c = d:FindContainerByName(container.name)
									c.options.HideButtonNoSenders = container.HideButtonNoSenders
									d:UpdateButtonVisuals(c, nil)
									d:SetContainerToVirtualSize(c)
								end,
						},
						HideButtonNoAvailableSenders = {
							type = "toggle",
							name = L["Hide Buttons without available Senders"],
							desc = L["Also hide buttons when all known Senders are dead, disconnected, or in a different zone."],
							disabled = function(info) return container.HideButtonNoSenders == false end,
							width = "full",
							get = function(info) return container.HideButtonNoAvailableSenders end,
							order = 3,
							set = 
								function(info, value)
									container.HideButtonNoAvailableSenders = value
									local d = self:GetModule("HermesDisplay")
									local c = d:FindContainerByName(container.name)
									c.options.HideButtonNoAvailableSenders = container.HideButtonNoAvailableSenders
									d:UpdateButtonVisuals(c, nil)
									d:SetContainerToVirtualSize(c)
								end,
						},
						ClassColoredBorder = {
							type = "toggle",
							name = L["Class Colored Borders"],
							desc = L["Sets the border color to match the class. Note that this effect is more apparent when using ButtonFacade to skin the buttons."],
							width = "full",
							get = function(info) return container.ClassColoredBorders end,
							order = 5,
							set = 
								function(info, value)
									container.ClassColoredBorders = value
									local d = self:GetModule("HermesDisplay")
									local c = d:FindContainerByName(container.name)
									c.options.ClassColoredBorders = container.ClassColoredBorders
									d:UpdateButtonBorders(c, nil)
								end,
						},
						Padding = {
							type = "range", min = 0, max = 100, step = 1,
							name = L["Padding"],
							width = "full",
							get = function(info) return container.padding end,
							order = 7,
							set = 
								function(info, value)
									container.padding = value
									local d = self:GetModule("HermesDisplay")
									local c = d:FindContainerByName(container.name)
									c.options.padding = container.padding
									d:ArrangeLayout(c)
									d:SetContainerToVirtualSize(c)
								end,
						},
					},
				},
				TicTacs = {
					name = L["Tic Tacs"],
					type = "group",
					inline = false,
					order = 2,
					args = {
						NumTicTacs = {
							type = "range",
							min = 0,
							max = self:GetModule("HermesDisplay").MAX_TICTACS,
							step = 1,
							name = L["Count"],
							order = 0,
							width = "full",
							get = function(info) return container.NumTicTacs end,
							set = 
								function(info, value)
									container.NumTicTacs = value
									local d = self:GetModule("HermesDisplay")
									local c = d:FindContainerByName(container.name)
									c.options.NumTicTacs = container.NumTicTacs
									d:ArrangeLayout(c)
									d:SetContainerToVirtualSize(c)
								end,
						},
						TicTacProperties = {
							name = " ",
							type = "group",
							inline = true,
							order = 1,
							args = {
								TicTacLocation= {
									type = "select",
									name = L["Location"],
									order = 1,
									style = "dropdown",
									width = "full",
									disabled = function() return container.NumTicTacs == 0 end,
									get = function(info)
										if(container.TicTacDirection == false and container.TicTacSide == false) then
											return "BOTTOM"
										elseif(container.TicTacDirection == false and container.TicTacSide == true) then
											return "TOP"
										elseif(container.TicTacDirection == true and container.TicTacSide == false) then
											return "RIGHT"
										elseif(container.TicTacDirection == true and container.TicTacSide == true) then
											return "LEFT"
										end
									end,
									set = 
										function(info, value)
											if(value == "BOTTOM") then
												container.TicTacDirection = false
												container.TicTacSide = false

											elseif(value == "TOP") then
												container.TicTacDirection = false
												container.TicTacSide = true

											elseif(value == "RIGHT") then
												container.TicTacDirection = true
												container.TicTacSide = false

											elseif(value == "LEFT") then
												container.TicTacDirection = true
												container.TicTacSide = true

											end

											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacDirection = container.TicTacDirection
											c.options.TicTacSide = container.TicTacSide
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
									values = {
										["RIGHT"] = L["Right"],
										["LEFT"] = L["Left"],
										["BOTTOM"] = L["Bottom"],
										["TOP"] = L["Top"],
									},
								},
								TicTacWidth = {
									type = "range",
									min = 1, max = 200, step = 1,
									name = L["Width"],
									width = "normal",
									get = function(info) return container.TicTacWidth end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 3,
									set = 
										function(info, value)
											container.TicTacWidth = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacWidth = container.TicTacWidth
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								TicTacHeight = {
									type = "range",
									min = 1, max = 200, step = 1,
									name = L["Height"],
									width = "normal",
									get = function(info) return container.TicTacHeight end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 5,
									set = 
										function(info, value)
											container.TicTacHeight = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacHeight = container.TicTacHeight
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								TicTacButtonOffset = {
									type = "range",
									min = 0, max = 20, step = 1,
									name = L["Offset"],
									width = "normal",
									get = function(info) return container.TicTacButtonOffset end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 8,
									set = 
										function(info, value)
											container.TicTacButtonOffset = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacButtonOffset = container.TicTacButtonOffset
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								TicTacGap = {
									type = "range",
									min = 0, max = 50, step = 1,
									name = L["Gap"],
									width = "normal",
									get = function(info) return container.TicTacGap end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 9,
									set = 
										function(info, value)
											container.TicTacGap = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacGap = container.TicTacGap
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								StatusIconSizeMultiplier = {
									type = "range",
									min = 0.1, max = 2, step = 0.01,
									name = L["Icon Size"],
									width = "normal",
									get = function(info) return container.StatusIconSizeMultiplier end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 10,
									set = 
										function(info, value)
											container.StatusIconSizeMultiplier = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.StatusIconSizeMultiplier = container.StatusIconSizeMultiplier
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								FontSizeMultiplier = {
									type = "range",
									min = 0.1, max = 2, step = 0.01,
									name = L["Font Size"],
									width = "normal",
									get = function(info) return container.PlayerNameSizeMultiplier end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 13,
									set = 
										function(info, value)
											container.PlayerNameSizeMultiplier = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.PlayerNameSizeMultiplier = container.PlayerNameSizeMultiplier
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								PlayerNameWidth = {
									type = "range",
									min = 1, max = 200, step = 1,
									name = L["Player Name Width"],
									width = "normal",
									get = function(info) return container.PlayerNameWidth end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 14,
									set = 
										function(info, value)
											container.PlayerNameWidth = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.PlayerNameWidth = container.PlayerNameWidth
											d:ArrangeLayout(c)
											d:SetContainerToVirtualSize(c)
										end,
								},
								TicTacGradientAlphaTop = {
									type = "range",
									min = 0, max = 1, step = 0.01,
									name = L["Top Gradient Effect"],
									width = "normal",
									get = function(info) return container.TicTacGradientAlphaTop end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 17,
									set = 
										function(info, value)
											container.TicTacGradientAlphaTop = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacGradientAlphaTop = container.TicTacGradientAlphaTop
											for _, button in ipairs(c.Buttons) do
												d:UpdateTicTacVisuals(button)
											end
										end,
								},
								TicTacGradientAlphaBottom = {
									type = "range",
									min = 0, max = 1, step = 0.01,
									name = L["Bottom Gradient Effect"],
									width = "normal",
									get = function(info) return container.TicTacGradientAlphaBottom end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 19,
									set = 
										function(info, value)
											container.TicTacGradientAlphaBottom = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacGradientAlphaBottom = container.TicTacGradientAlphaBottom
											for _, button in ipairs(c.Buttons) do
												d:UpdateTicTacVisuals(button)
											end
										end,
								},
								TicTacGradientAlphaNoPlayer = {
									type = "range",
									min = 0, max = 1, step = 0.01,
									name = L["No Sender Alpha"],
									width = "normal",
									get = function(info) return container.TicTacGradientAlphaNoPlayer end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 21,
									set = 
										function(info, value)
											container.TicTacGradientAlphaNoPlayer = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacGradientAlphaNoPlayer = container.TicTacGradientAlphaNoPlayer
											for _, button in ipairs(c.Buttons) do
												d:UpdateTicTacVisuals(button)
											end
										end,
								},
								TicTacGradientAlphaNotAvailable = {
									type = "range",
									min = 0, max = 1, step = 0.01,
									name = L["Not Available Alpha"],
									width = "normal",
									get = function(info) return container.TicTacGradientAlphaNotAvailable end,
									disabled = function() return container.NumTicTacs == 0 end,
									order = 21,
									set = 
										function(info, value)
											container.TicTacGradientAlphaNotAvailable = value
											local d = self:GetModule("HermesDisplay")
											local c = d:FindContainerByName(container.name)
											c.options.TicTacGradientAlphaNotAvailable = container.TicTacGradientAlphaNotAvailable
											for _, button in ipairs(c.Buttons) do
												d:UpdateTicTacVisuals(button)
											end
										end,
								},
							},
						},
					},
				},
			}
		}
	end
	
	optionTables.args.Containers.args.Common = {
		type="group",
		name=" ",
		inline = true,
		order = 0,
		args = {
			LockContainers = {
				type = "toggle",
				name = L["Lock Containers"],
				width = "normal",
				--disabled = function() return self.db.profile.enabled == false or self.db.profile.configMode == true end,
				disabled = function() return self.db.profile.enabled == false end,
				get = function(info) return self.db.profile.locked end,
				order = 1,
				set =
					function(info, value)
						self.db.profile.locked = value
						local d = self:GetModule("HermesDisplay")
						self:GetModule("HermesDisplay"):UpdateContainerLocking()
					end,
			},
			Add = {
				type = "execute",
				name = L["Add"],
				width = "normal",
				order = 2,
				func = "OnAddContainer",
			}
		}
	}
end

function Hermes:OnSetEnabled(info, value)
	self.db.profile.enabled = value
	if(value == true) then
		Hermes:Startup()
	else
		Hermes:Shutdown()
	end
end

function Hermes:OnSetConfigMode(info, value)
	self.db.profile.configMode = value
	local d = self:GetModule("HermesDisplay")
	--make sure windows are visible or not based on config mode and other properties
	self:GetModule("HermesDisplay"):UpdateContainerVisibility(self.db.profile.enabled, self.db.profile.receiver.enabled and Hermes.Receiver ~= nil)
	--lock containers based on config mode
	self:GetModule("HermesDisplay"):UpdateContainerLocking()
end

function Hermes:OnSetEnablePartySupport(info, value)
	self.db.profile.enableparty = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetEnableSender(info, value)
	self.db.profile.sender.enabled = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetEnableReceiver(info, value)
	self.db.profile.receiver.enabled = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnAddContainer()
	local display = self:GetModule("HermesDisplay")
	--generate a unique name for the container
	local name = GenerateUniqueContainerName(display.Containers, L["Container"].." ")
	--clone defaults for the container
	local db = deepcopy(newContainerOptions)
	--make sure to set the container name!!
	db.name = name
	--add container db values to settings
	tinsert(Hermes.db.profile.containers, db)
	--add container to display, need to copy the defaults again due to profile changes
	display:AddContainer(deepcopy(db))

	--unlock windows, pure convenience...
	Hermes.db.profile.locked = false
	display:UpdateContainerLocking()
	
	local container = display:FindContainerByName(name)
	display:SaveContainerPosition(container) --required as this ends up calling UpdateContainerProfilePosition which sets the X and Y in the profile. Without it the reload of the options tble will result in error where it tries to Round x and y
	
	--force the config window to update with latest container added
	self:BlizOptionsTable_Containers()
	LIB_AceConfigDialog:SelectGroup(AddonName, "Containers", name)
end

function Hermes:OnAddSpell(info)
	--see if the spell exists
	local name
	local icon
	if(newSpellTemplate.id) then
		name, _, icon, _, _, _, _, _, _ = GetSpellInfo(newSpellTemplate.id)
	else
		--try to find the spell id for the given spell name
		local matches = self:GetSpellID(newSpellTemplate.name)
		
		if(matches and #matches > 0) then
			--still need the icon
			newSpellTemplate.id = matches[1]
			name, _, icon, _, _, _, _, _, _ = GetSpellInfo(newSpellTemplate.id)
			if(#matches > 1) then
				--there are more than one spellid from the name given, ask the user which they want to use
				local message = L["|cFFFF0000Hermes Warning|r"].." "..L["multiple id's were found. The first id was chosen"]
				print(message)
			end
		end
	end
	
	if(name and icon) then
		--make sure we're not already tracking this spell
		local id = self:EncodeTrackerIdentifier(tonumber(newSpellTemplate.id), "spell")
		for i, spell in ipairs(self.db.profile.spells) do
			if(spell.id == id) then
				local message = L["|cFFFF0000Hermes Warning|r"].." "..L["spell has already been added"]
				print(message)
				return
			end
		end
		
		local spell = {
			class = newSpellTemplate.class,
			id = tonumber(id),
			name = name,
			icon = icon,
			enabled = true,
		}
		
		--add to db
		tinsert(self.db.profile.spells, spell)
		
		--sort the spells
		table.sort(self.db.profile.spells, function(a, b) return Hermes:SortProfileSpells(a, b) end)

		--add to display
		local HermesDisplay = self:GetModule("HermesDisplay")
		HermesDisplay:AddTracker(spell.id, spell.name, spell.icon, nil, spell.class)
		--end
		
		--reset the template, but leave the class alone
		newSpellTemplate.id = nil
		newSpellTemplate.name = nil
		
		--request spell from senders
		if(Hermes:IsReceiving()) then
			self:SendMessage_REQUEST_SPELLS(nil, {spell.id})
		end
	
		--force the config window to update with latest spell info
		LIB_AceConfigRegistry:NotifyChange(AddonName)
		Hermes:BlizOptionsTable_Spells()
	else
		local message = L["|cFFFF0000Hermes Warning|r"].." "..L["spell name or id not found"]
		print(message)
	end
end

function Hermes:OnAddItem(info)
	local id, name, icon = self:GetItemInfo(newItemTemplate.id, newItemTemplate.name)
	if(not icon) then
		local message = L["|cFFFF0000Hermes Warning|r"].." "..L["Item name or id not found. If you're confident the id or name is correct, try having someone link the item or putting the item in your inventory and adding again."]
		print(message)
		return
	end
	
	if(icon and id) then
		id = self:EncodeTrackerIdentifier(tonumber(id), "item")
		--make sure we're not already tracking this item
		for i, item in ipairs(self.db.profile.items) do
			if(item.id == tonumber(id)) then
				local message = L["|cFFFF0000Hermes Warning|r"].." "..L["Item has already been added"]
				print(message)
				return
			end
		end
		
		--we need to cache the item
		local caching = nil
		if(not name) then
			caching = 1
			name  = "..."..L["searching"].." ("..tostring(self:DecodeTrackerIdentifier(tonumber(id)))..")"
		end
		
		local item = {
			class = newItemTemplate.class,
			id = tonumber(id),
			name = name,
			icon = icon,
			enabled = true,
			caching = caching
		}
		
		--add to db
		tinsert(self.db.profile.items, item)
		
		--sort the spells
		table.sort(self.db.profile.items, function(a, b) return Hermes:SortProfileItems(a, b) end)

		--add to display
		local HermesDisplay = self:GetModule("HermesDisplay")
		HermesDisplay:AddTracker(item.id, item.name, item.icon, nil, item.class)
		--end
		
		--reset the template, but leave the class alone
		newItemTemplate.id = nil
		newItemTemplate.name = nil
		
		--request spell from senders
		if(Hermes:IsReceiving()) then
			self:SendMessage_REQUEST_SPELLS(nil, {item.id})
		end
	
		--force the config window to update with latest spell info
		LIB_AceConfigRegistry:NotifyChange(AddonName)
		Hermes:BlizOptionsTable_Items()
		
		--fire off the name cache timer if needed
		if(caching) then
			Hermes:StartItemNameTimer(item.id)
		end
	end
end

function Hermes:OnDeleteSpell(info)
	local spellid = info[#info - 1]
	local classToken = info[#info - 2]
	
	--locate the index of the spell in the db
	local index = nil
	for i, spell in ipairs(self.db.profile.spells) do
		if(tostring(spell.id) == spellid and spell.class == classToken) then
			index = i
			break
		end
	end
	
	if(not index) then
		error("Unexpected error trying to delete spell from profile. Data consistency issue could mean more problems")
	end
	
	--delete from profile
	tremove(self.db.profile.spells, index)
	
	local HermesDisplay = self:GetModule("HermesDisplay")
	HermesDisplay:RemoveTracker(tonumber(spellid))
	
	--force the config window to update with latest spell info
	LIB_AceConfigRegistry:NotifyChange(AddonName)
	Hermes:BlizOptionsTable_Spells()
end

function Hermes:OnDeleteItem(info)
	local itemid = info[#info - 1]
	local classToken = info[#info - 2]
	
	--locate the index of the spell in the db
	local index = nil
	for i, item in ipairs(self.db.profile.items) do
		if(tostring(item.id) == itemid and item.class == classToken) then
			index = i
			break
		end
	end
	
	if(not index) then
		error("Unexpected error trying to delete item from profile. Data consistency issue could mean more problems")
	end
	
	--delete from profile
	tremove(self.db.profile.items, index)
	
	local HermesDisplay = self:GetModule("HermesDisplay")
	HermesDisplay:RemoveTracker(tonumber(itemid))
	
	--force the config window to update with latest spell info
	LIB_AceConfigRegistry:NotifyChange(AddonName)
	Hermes:BlizOptionsTable_Items()
end

function Hermes:OnDeleteContainer(info)
	local containerName = info[#info - 3]
	local HermesDisplay = self:GetModule("HermesDisplay")
	local container = HermesDisplay:FindContainerByName(containerName)
	
	if(container.options.name == nil) then
		error("Attempt to delete Default frame, should not have been possible")
	end
	
	--move any buttons to default container
	while #container.Buttons > 0 do
		HermesDisplay:MoveButtonBetweenContainers(container.Buttons[1], HermesDisplay.DefaultContainer) --move all buttons in container to nil container
	end
	
	--now we need to update any remaining spells and item in the profile that weren't in the container but are assigned to the container (they are not enabled)
	for _, spell in ipairs(self.db.profile.spells) do
		if(spell.container == container.options.name) then
			--no need to move it around because presumably it's not enabled anyway. Throw this check in there to catch data integrity/coding errors
			if(spell.enabled == true) then error("unexpected condition encountered looking for spell containers") end
			spell.container = nil
		end
		if(spell.priorcontainer == container.options.name) then
			spell.priorcontainer = nil
		end
	end
	
	for _, item in ipairs(self.db.profile.items) do
		if(item.container == container.options.name) then
			--no need to move it around because presumably it's not enabled anyway. Throw this check in there to catch data integrity/coding errors
			if(item.enabled == true) then error("unexpected condition encountered looking for item containers") end
			item.container = nil
		end
		if(item.priorcontainer == container.options.name) then
			item.priorcontainer = nil
		end
	end
	
	--recycle the container
	HermesDisplay:RecycleContainer(container)
	
	--remove the container from db
	local index = nil
	for i, c in ipairs(self.db.profile.containers) do
		if(c.name and c.name == containerName) then
			index = i
			break
		end
	end
	
	if(not index) then
		error("Unexpected error trying to delete container from profile. Data consistency issue could mean more problems")
	end
	
	--delete from profile
	tremove(self.db.profile.containers, index)
	
	--update config display
	LIB_AceConfigRegistry:NotifyChange(AddonName)
	Hermes:BlizOptionsTable_Containers()
end

function Hermes:OnResetContainerPosition(info)
	local containerName = info[#info - 3]
	local HermesDisplay = self:GetModule("HermesDisplay")
	local container = HermesDisplay:FindContainerByName(containerName)
	container:ClearAllPoints()
	container:SetPoint("CENTER", UIParent, "CENTER")
	container.options.x = container:GetLeft()
	container.options.y = container:GetTop()
	HermesDisplay:SaveContainerPosition(container)
end
